import drjit as dr
import pytest
import importlib


def test01_init_zero():
    from drjit.scalar import Array0f, Array3f, ArrayXf
    a = Array0f()
    assert len(a) == 0

    a = Array3f()
    assert len(a) == 3
    for i in range(3):
        if dr.DEBUG:
            assert a[i] != a[i]
        else:
            assert a[i] == 0

    a = ArrayXf()
    assert len(a) == 0
    with pytest.raises(IndexError) as ei:
        a[0]
    assert "entry 0 is out of bounds (the array is of size 0)." in str(ei.value)


def test02_init_sequence_static():
    from drjit.scalar import Array0f, Array1f, Array3f

    a = Array3f((0, 1, 2))
    for i in range(3):
        assert a[i] == i

    a = Array3f([0, 1, 2])
    for i in range(3):
        assert a[i] == i

    class my_list(list):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    a = Array3f(my_list([0, 1, 2]))
    for i in range(3):
        assert a[i] == i

    a = Array3f(0, 1, 2)
    for i in range(3):
        assert a[i] == i

    assert Array1f(1)[100] == 1

    pytest.skip("nanobind layer")

    with pytest.raises(TypeError) as ei:
        Array3f("test")
    assert "input sequence has wrong size (expected 3, got 4)" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        Array0f(1)
    assert "too many arguments provided" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        a = Array3f((0, 1))
    assert "input sequence has wrong size" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        a = Array3f((0, 1, 2, 3))
    assert "input sequence has wrong size" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        Array3f("tst")
    assert "could not initialize element with a value of type 'str'" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        Array3f((0, "foo", 2))
    assert "could not initialize element with a value of type 'str'" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        Array3f(0, "foo", 2)
    assert "could not initialize element with a value of type 'str'" in str(ei.value)


def test03_init_sequence_dynamic():
    from drjit.scalar import ArrayXf

    a = ArrayXf((0, 1, 2, 3, 4))
    assert len(a) == 5
    for i in range(5):
        assert a[i] == i

    a = ArrayXf([0, 1, 2])
    assert len(a) == 3
    for i in range(3):
        assert a[i] == i

    class my_list(list):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    a = ArrayXf(my_list([0, 1, 2]))
    for i in range(3):
        assert a[i] == i

    a = ArrayXf(0, 1, 2)
    for i in range(3):
        assert a[i] == i

    for i in range(3):
        a[i] += 1

    for i in range(3):
        assert a[i] == i + 1

    assert ArrayXf(1)[100] == 1

    pytest.skip("nanobind layer")

    with pytest.raises(TypeError) as ei:
        ArrayXf("test")
    assert "could not initialize element with a value of type 'str'" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        ArrayXf((0, "foo", 2))
    assert "could not initialize element with a value of type 'str'" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        ArrayXf(0, "foo", 2)
    assert "could not initialize element with a value of type 'str'" in str(ei.value)



def test04_indexing():
    from drjit.scalar import ArrayXf, Array3f
    a = ArrayXf([0, 1, 2])
    b = Array3f([0, 1, 2])
    assert a[-1] == 2 and b[-1] == 2

    with pytest.raises(IndexError) as ei:
        a[3]

    with pytest.raises(IndexError) as ei:
        a[-4]

    with pytest.raises(IndexError) as ei:
        b[3]

    with pytest.raises(IndexError) as ei:
        b[-4]


def test05_indexed_assignment():
    from drjit.scalar import ArrayXf, Array3f

    a = Array3f([0]*3)
    b = ArrayXf([0]*5)
    for i in range(3):
        a[i] = i
    for i in range(len(a)):
        a[i] += 1
    for i in range(len(b)):
        b[i] = i
    for i in range(len(b)):
        b[i] += 1
    for i in range(len(a)):
        assert a[i] == i + 1
    for i in range(len(b)):
        assert b[i] == i + 1


def test06_constructor_copy():
    from drjit.scalar import Array3f, ArrayXf
    a = Array3f(1, 2, 3)
    b = Array3f(a)
    c = ArrayXf(1, 2, 3, 4)
    d = ArrayXf(c)
    assert len(a) == len(b)
    assert len(c) == len(d)
    for i in range(len(a)):
        assert a[i] == b[i]
    for i in range(len(c)):
        assert c[i] == d[i]


def test07_constructor_broadcast():
    from drjit.scalar import Array3f, ArrayXf, ArrayXb
    a = Array3f(3)
    assert len(a) == 3 and a[0] == 3 and a[1] == 3 and a[2] == 3
    a = ArrayXf(3)
    assert len(a) == 1 and a[0] == 3
    a = ArrayXb(True)
    assert len(a) == 1 and a[0] == True


def test08_all_any():
    from drjit.scalar import Array2b, ArrayXb

    assert dr.all(True) == True
    assert dr.all(False) == False
    assert dr.any(True) == True
    assert dr.any(False) == False
    assert dr.any(()) == False
    assert dr.all(()) == True
    assert dr.all((True,)) == True
    assert dr.all((False,)) == False
    assert dr.any((True,)) == True
    assert dr.any((False,)) == False
    assert dr.all([True, True]) == True
    assert dr.all([True, False]) == False
    assert dr.all([False, False]) == False
    assert dr.any([True, True]) == True
    assert dr.any([True, False]) == True
    assert dr.any([False, False]) == False
    assert type(dr.all(Array2b(True, True))) is bool
    assert dr.all(Array2b(True, True)) == True
    assert dr.all(Array2b(True, True)) == True
    assert dr.all(Array2b(True, False)) == False
    assert dr.all(Array2b(False, False)) == False
    assert dr.any(Array2b(True, True)) == True
    assert dr.any(Array2b(True, False)) == True
    assert dr.any(Array2b(False, False)) == False

    pytest.skip("nanobind layer")

    assert type(dr.all(ArrayXb(True, True))) is ArrayXb
    assert len(dr.all(ArrayXb(True, True))) == 1
    assert dr.all(ArrayXb(True, True))[0] == True
    assert dr.all(ArrayXb(True, False))[0] == False
    assert dr.all(ArrayXb(False, False))[0] == False
    assert dr.any(ArrayXb(True, True))[0] == True
    assert dr.any(ArrayXb(True, False))[0] == True
    assert dr.any(ArrayXb(False, False))[0] == False

    assert type(dr.all(dr.llvm.Array1b(dr.llvm.Bool([True, False, False])))) is dr.llvm.Bool
    assert len(dr.all(dr.llvm.Array1b(dr.llvm.Bool([True, False, False])))) == 3
    assert type(dr.all_nested(dr.llvm.Array1b(dr.llvm.Bool([True, False, False])))) is dr.llvm.Bool
    assert len(dr.all_nested(dr.llvm.Array1b(dr.llvm.Bool([True, False, False])))) == 1

    with pytest.raises(TypeError) as ei:
        dr.all((True, "hello"))
    assert "unsupported operand type(s)" in str(ei.value)

@pytest.mark.skip("nanobind layer")
def test09_implicit_to_bool():
    from drjit.scalar import Array3f, ArrayXf, Array3b, ArrayXb
    with pytest.raises(TypeError) as ei:
        bool(ArrayXf(1))
    assert "ArrayXf.__bool__(): implicit conversion to 'bool' is only supported for scalar mask arrays!" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        bool(Array3f(1))
    assert "Array3f.__bool__(): implicit conversion to 'bool' is only supported for scalar mask arrays!" in str(ei.value)

    with pytest.raises(RuntimeError) as ei:
        bool(ArrayXb(True, False))
    assert "ArrayXb.__bool__(): implicit conversion to 'bool' requires a scalar mask array (array size was 2)." in str(ei.value)

    assert bool(ArrayXb(True))
    assert not bool(ArrayXb(False))


@pytest.mark.skip("nanobind layer")
@pytest.mark.parametrize('value', [(dr.scalar.Array3f, dr.scalar.Array3b), (dr.scalar.ArrayXf, dr.scalar.ArrayXb)])
def test10_comparison(value):
    c, m = value
    assert dr.all(c(1, 2, 3) == c(1, 2, 3))
    assert not dr.all(c(1, 2, 3) == c(1, 3, 3))
    assert dr.all(c(1, 2, 3) != c(4, 5, 6))
    assert not dr.all(c(1, 2, 3) != c(4, 2, 6))
    assert dr.any(c(1, 2, 3) == c(1, 2, 3))
    assert not dr.any(c(1, 2, 3) == c(4, 5, 6))
    assert dr.any(c(1, 2, 3) != c(1, 3, 3))
    assert not dr.any(c(1, 2, 3) != c(1, 2, 3))

    assert dr.all((c(1, 2, 3) < c(0, 2, 4)) == m(False, False, True))
    assert dr.all((c(1, 2, 3) <= c(0, 2, 4)) == m(False, True, True))
    assert dr.all((c(1, 2, 3) > c(0, 2, 4)) == m(True, False, False))
    assert dr.all((c(1, 2, 3) >= c(0, 2, 4)) == m(True, True, False))
    assert dr.all((c(1, 2, 3) == c(0, 2, 4)) == m(False, True, False))
    assert dr.all((c(1, 2, 3) != c(0, 2, 4)) == m(True, False, True))


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test11_shape():
    import drjit.scalar as s
    import drjit.llvm as l

    assert dr.shape(s.Array0f()) == (0,) and s.Array0f().shape == (0,)
    assert dr.shape(s.Array2f()) == (2,) and s.Array2f().shape == (2,)
    assert dr.shape(l.Float()) == (0,) and l.Float().shape == (0,)
    assert dr.shape(l.Float(1, 2, 3)) == (3,) and l.Float(1, 2, 3).shape == (3,)
    assert dr.shape(l.Array2f()) == (2, 0) and l.Array2f().shape == (2, 0)
    assert dr.shape(l.Array2f(l.Float(1, 2, 3))) == (2,3) and \
           l.Array2f(l.Float(1, 2, 3)).shape == (2,3)
    assert dr.shape(l.Array2f(l.Float(1, 2, 3),
                              l.Float(2, 3, 4))) == (2,3) and \
            l.Array2f(l.Float(1, 2, 3), l.Float(2, 3, 4)).shape == (2,3)
    assert dr.shape(l.Array2f(l.Float(1, 2, 3),
                              l.Float(2, 3))) is None and \
           l.Array2f(l.Float(1, 2, 3), l.Float(2, 3)).shape is None

@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test11_repr():
  import drjit.scalar as s
  import drjit.llvm as l

  assert repr(s.Array0f()) == '[]'
  assert repr(s.ArrayXf()) == '[]'
  assert repr(s.Array1f(1)) == '[1]'
  assert repr(s.Array1f(1.5)) == '[1.5]'
  assert repr(s.Array2f(1, 2)) == '[1, 2]'
  assert repr(s.ArrayXf(1, 2)) == '[1, 2]'

  assert repr(l.Array0f()) == '[]'
  assert repr(l.ArrayXf()) == '[]'
  assert repr(l.Array1f(1)) == '[[1]]'
  assert repr(l.Array1f(1.5)) == '[[1.5]]'
  assert repr(l.Array2f(1, 2)) == '[[1, 2]]'
  assert repr(l.Array2f(1, [2, 3])) == '[[1, 2],\n' \
                                       ' [1, 3]]'
  assert repr(l.ArrayXf(1, 2)) == '[[1, 2]]'
  assert repr(l.Float(range(1000))) == '[0, 1, 2, 3, 4, .. 990 skipped .., 995, 996, 997, 998, 999]'


def test12_binop_simple():
    from drjit.scalar import Array3f, ArrayXf, Array3u, ArrayXb
    a = Array3f(1, 2, 3)
    assert dr.all(a + a == Array3f(2, 4, 6))
    assert dr.all(a + (1, 2, 3) == Array3f(2, 4, 6))
    assert dr.all(a + [1, 2, 3] == Array3f(2, 4, 6))
    assert dr.all((1, 2, 3) + a == Array3f(2, 4, 6))
    assert dr.all([1, 2, 3] + a == Array3f(2, 4, 6))
    assert dr.all(a - a == Array3f(0, 0, 0))
    assert dr.all(a * a == Array3f(1, 4, 9))
    assert dr.all(a / a == Array3f(1, 1, 1))

    with pytest.raises(TypeError) as ei:
        a // a
    # assert "unsupported operand type(s)" in str(ei.value) # TODO

    a = ArrayXf(1, 2, 3)
    assert dr.all(a + a == ArrayXf(2, 4, 6))
    assert dr.all(a - a == ArrayXf(0, 0, 0))
    assert dr.all(a * a == ArrayXf(1, 4, 9))
    assert dr.all(a / a == ArrayXf(1, 1, 1))

    a = Array3u(1, 2, 3)
    assert dr.all(a + a == Array3u(2, 4, 6))
    assert dr.all(a - a == Array3u(0, 0, 0))
    assert dr.all(a * a == Array3u(1, 4, 9))
    assert dr.all(a // a == Array3u(1, 1, 1))
    assert dr.all(a << 1 == Array3u(2, 4, 6))
    assert dr.all(a >> 1 == Array3u(0, 1, 1))

    with pytest.raises(TypeError) as ei:
        a / a
    # assert "unsupported operand type(s)" in str(ei.value) # TODO

    assert dr.all(ArrayXb([True, True, False, False]) & ArrayXb([True, False, True, False]) == ArrayXb(True, False, False, False))
    assert dr.all(ArrayXb([True, True, False, False]) | ArrayXb([True, False, True, False]) == ArrayXb(True, True, True, False))
    assert dr.all(ArrayXb([True, True, False, False]) ^ ArrayXb([True, False, True, False]) == ArrayXb(False, True, True, False))


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test13_binop_promote_broadcast():
    import drjit.llvm as l
    import drjit.scalar as s

    x = s.ArrayXf(10, 100, 1000) + 1
    assert type(x) is s.ArrayXf and dr.all(x == s.ArrayXf(11, 101, 1001))
    x = 1 + s.ArrayXf(10, 100, 1000)
    assert type(x) is s.ArrayXf and dr.all(x == s.ArrayXf(11, 101, 1001))
    x = s.ArrayXf(10, 100, 1000) + (1, 2, 3)
    assert type(x) is s.ArrayXf and dr.all(x == s.ArrayXf(11, 102, 1003))
    x = (1, 2, 3) + s.ArrayXf(10, 100, 1000)
    assert type(x) is s.ArrayXf and dr.all(x == s.ArrayXf(11, 102, 1003))
    x = [1, 2, 3] + s.ArrayXf(10, 100, 1000)
    assert type(x) is s.ArrayXf and dr.all(x == s.ArrayXf(11, 102, 1003))
    x = s.ArrayXf(10, 100, 1000) + [1, 2, 3]
    assert type(x) is s.ArrayXf and dr.all(x == s.ArrayXf(11, 102, 1003))

    x = s.Array3f(10, 100, 1000) + 1
    assert type(x) is s.Array3f and dr.all(x == s.Array3f(11, 101, 1001))
    x = 1 + s.Array3f(10, 100, 1000)
    assert type(x) is s.Array3f and dr.all(x == s.Array3f(11, 101, 1001))
    x = s.Array3f(10, 100, 1000) + (1, 2, 3)
    assert type(x) is s.Array3f and dr.all(x == s.Array3f(11, 102, 1003))
    x = (1, 2, 3) + s.Array3f(10, 100, 1000)
    assert type(x) is s.Array3f and dr.all(x == s.Array3f(11, 102, 1003))
    x = [1, 2, 3] + s.Array3f(10, 100, 1000)
    assert type(x) is s.Array3f and dr.all(x == s.Array3f(11, 102, 1003))
    x = s.Array3f(10, 100, 1000) + [1, 2, 3]
    assert type(x) is s.Array3f and dr.all(x == s.Array3f(11, 102, 1003))

    x = l.Float(10, 100, 1000) + 1
    assert type(x) is l.Float and dr.all(x == l.Float(11, 101, 1001))
    x = 1 + l.Float(10, 100, 1000)
    assert type(x) is l.Float and dr.all(x == l.Float(11, 101, 1001))
    x = l.Float(10, 100, 1000) + (1, 2, 3)
    assert type(x) is l.Float and dr.all(x == l.Float(11, 102, 1003))
    x = (1, 2, 3) + l.Float(10, 100, 1000)
    assert type(x) is l.Float and dr.all(x == l.Float(11, 102, 1003))
    x = [1, 2, 3] + l.Float(10, 100, 1000)
    assert type(x) is l.Float and dr.all(x == l.Float(11, 102, 1003))
    x = l.Float(10, 100, 1000) + [1, 2, 3]
    assert type(x) is l.Float and dr.all(x == l.Float(11, 102, 1003))

    x = l.Array3f(10, 100, 1000) + 1
    assert type(x) is l.Array3f and dr.all_nested(x == l.Array3f(11, 101, 1001))
    x = 1 + l.Array3f(10, 100, 1000)
    assert type(x) is l.Array3f and dr.all_nested(x == l.Array3f(11, 101, 1001))
    x = l.Array3f(10, 100, 1000) + (1, 2, 3)
    assert type(x) is l.Array3f and dr.all_nested(x == l.Array3f(11, 102, 1003))
    x = (1, 2, 3) + l.Array3f(10, 100, 1000)
    assert type(x) is l.Array3f and dr.all_nested(x == l.Array3f(11, 102, 1003))
    x = [1, 2, 3] + l.Array3f(10, 100, 1000)
    assert type(x) is l.Array3f and dr.all_nested(x == l.Array3f(11, 102, 1003))
    x = l.Array3f(10, 100, 1000) + [1, 2, 3]
    assert type(x) is l.Array3f and dr.all_nested(x == l.Array3f(11, 102, 1003))

    x = s.Array3i(10, 100, 1000) + l.Float(1)
    assert type(x) is l.Array3f and dr.all_nested(x == l.Array3f(11, 101, 1001))
    x = s.Array3f(10, 100, 1000) + l.Float(1)
    assert type(x) is l.Array3f and dr.all_nested(x == l.Array3f(11, 101, 1001))


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test14_binop_inplace():
    import drjit.scalar as s
    import drjit.llvm as l

    a = s.Array3f(1, 2, 3)
    b = s.Array3f(2, 3, 1)
    c = a
    a += b
    assert a is c and dr.all(a == s.Array3f(3, 5, 4))
    a += 1
    assert a is c and dr.all(a == s.Array3f(4, 6, 5))
    a = 1
    c = a
    a += b
    assert a is not c and dr.all(a == s.Array3f(3, 4, 2))

    a = s.ArrayXf(1, 2, 3)
    b = s.ArrayXf(2, 3, 1)
    c = a
    a += b
    assert a is c and dr.all(a == s.ArrayXf(3, 5, 4))
    a += 1
    assert a is c and dr.all(a == s.ArrayXf(4, 6, 5))
    a = 1
    c = a
    a += b
    assert a is not c and dr.all(a == s.ArrayXf(3, 4, 2))

    a = l.Float(1, 2, 3)
    b = l.Float(2, 3, 1)
    c = a
    a += b
    assert a is c and dr.all(a == l.Float(3, 5, 4))
    a += 1
    assert a is c and dr.all(a == l.Float(4, 6, 5))
    a = 1
    c = a
    a += b
    assert a is not c and dr.all(a == l.Float(3, 4, 2))

    a = l.ArrayXf(1, 2, 3)
    b = l.ArrayXf(2, 3, 1)
    c = a
    a += b
    assert a is c and dr.all(a == l.ArrayXf(3, 5, 4))
    a = 1
    c = a
    a += b
    assert a is not c and dr.all(a == l.ArrayXf(3, 4, 2))


@pytest.mark.skip("nanobind layer")
@pytest.mark.parametrize('m', [dr.scalar, dr.llvm])
def test15_unop(m):
    assert dr.all(-m.ArrayXf(1, 2, 3) == m.ArrayXf(-1, -2, -3))
    assert dr.all(+m.ArrayXf(1, 2, 3) == m.ArrayXf(1, 2, 3))
    assert dr.all(abs(m.ArrayXf(1, -2, 3)) == m.ArrayXf(1, 2, 3))
    assert dr.all(-m.Array3f(1, 2, 3) == m.Array3f(-1, -2, -3))
    assert dr.all(+m.Array3f(1, 2, 3) == m.Array3f(1, 2, 3))
    assert dr.all(abs(m.Array3f(1, -2, 3)) == m.Array3f(1, 2, 3))
    assert dr.all(-m.Array3i(1, 2, 3) == m.Array3i(-1, -2, -3))
    assert dr.all(+m.Array3i(1, 2, 3) == m.Array3i(1, 2, 3))
    assert dr.all(~m.Array3i(1, 2, 3) == m.Array3i(-2, -3, -4))
    assert dr.all(abs(m.Array3i(1, -2, 3)) == m.Array3i(1, 2, 3))
    assert dr.all(~m.Array3b(True, False, True) == m.Array3b(False, True, False))
    assert dr.all(~m.ArrayXb(True, False, True) == m.ArrayXb(False, True, False))


@pytest.mark.skip("nanobind layer")
def test16_type_promotion_errors():
    from drjit.scalar import Array3f
    a = Array3f()
    with pytest.raises(TypeError) as ei:
        a + "asdf"
    assert "Array3f.__add__(): encountered an unsupported argument of type 'str' (must be a Dr.Jit array or a type that can be converted into one)" in str(ei.value)

    a + 2**10
    with pytest.raises(TypeError):
        a + 2**100


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test18_traits():
    import drjit.scalar as s
    import drjit.llvm as l
    from drjit import Dynamic

    assert not dr.is_array_v(()) and not dr.is_array_v(1.0)
    assert dr.is_array_v(s.Array3f) and dr.is_array_v(s.Array3f())
    assert dr.is_array_v(s.ArrayXf) and dr.is_array_v(s.ArrayXf())
    assert dr.is_array_v(l.Array3f) and dr.is_array_v(l.Array3f())
    assert dr.is_array_v(l.ArrayXf) and dr.is_array_v(l.ArrayXf())
    assert dr.size_v(1) == 1
    assert dr.size_v("test") == 1
    assert dr.size_v(s.Array3f) == 3 and dr.size_v(s.Array3f()) == 3
    assert dr.size_v(l.Array3f) == 3 and dr.size_v(l.Array3f()) == 3
    assert dr.size_v(s.ArrayXf) == Dynamic and dr.size_v(s.ArrayXf()) == Dynamic
    assert dr.size_v(l.ArrayXf) == Dynamic and dr.size_v(l.ArrayXf()) == Dynamic

    assert dr.depth_v(1) == 0
    assert dr.depth_v("test") == 0
    assert dr.depth_v(s.Array3f) == 1 and dr.depth_v(s.Array3f()) == 1
    assert dr.depth_v(s.ArrayXf) == 1 and dr.depth_v(s.ArrayXf()) == 1
    assert dr.depth_v(l.Float) == 1 and dr.depth_v(l.Float()) == 1
    assert dr.depth_v(l.Array3f) == 2 and dr.depth_v(l.Array3f()) == 2
    assert dr.depth_v(l.ArrayXf) == 2 and dr.depth_v(l.ArrayXf()) == 2

    assert dr.scalar_t(1) is int
    assert dr.scalar_t("test") is str
    assert dr.scalar_t(s.Array3f) is float and dr.scalar_t(s.Array3f()) is float
    assert dr.scalar_t(s.Array3b) is bool and dr.scalar_t(s.Array3b()) is bool
    assert dr.scalar_t(s.ArrayXf) is float and dr.scalar_t(s.ArrayXf()) is float
    assert dr.scalar_t(s.ArrayXb) is bool and dr.scalar_t(s.ArrayXb()) is bool
    assert dr.scalar_t(l.Float) is float and dr.scalar_t(l.Float()) is float
    assert dr.scalar_t(l.Bool) is bool and dr.scalar_t(l.Bool()) is bool
    assert dr.scalar_t(l.Array3f) is float and dr.scalar_t(l.Array3f()) is float
    assert dr.scalar_t(l.Array3b) is bool and dr.scalar_t(l.Array3b()) is bool
    assert dr.scalar_t(l.ArrayXf) is float and dr.scalar_t(l.ArrayXf()) is float
    assert dr.scalar_t(l.ArrayXb) is bool and dr.scalar_t(l.ArrayXb()) is bool

    assert dr.value_t(1) is int
    assert dr.value_t("test") is str
    assert dr.value_t(s.Array3f) is float and dr.value_t(s.Array3f()) is float
    assert dr.value_t(s.Array3b) is bool and dr.value_t(s.Array3b()) is bool
    assert dr.value_t(s.ArrayXf) is float and dr.value_t(s.ArrayXf()) is float
    assert dr.value_t(s.ArrayXb) is bool and dr.value_t(s.ArrayXb()) is bool
    assert dr.value_t(l.Float) is float and dr.value_t(l.Float()) is float
    assert dr.value_t(l.Bool) is bool and dr.value_t(l.Bool()) is bool
    assert dr.value_t(l.Array3f) is l.Float and dr.value_t(l.Array3f()) is l.Float
    assert dr.value_t(l.Array3b) is l.Bool and dr.value_t(l.Array3b()) is l.Bool
    assert dr.value_t(l.ArrayXf) is l.Float and dr.value_t(l.ArrayXf()) is l.Float
    assert dr.value_t(l.ArrayXb) is l.Bool and dr.value_t(l.ArrayXb()) is l.Bool

    assert dr.mask_t(1) is bool
    assert dr.mask_t("test") is bool
    assert dr.mask_t(s.Array3f) is s.Array3b and dr.mask_t(s.Array3f()) is s.Array3b
    assert dr.mask_t(s.Array3b) is s.Array3b and dr.mask_t(s.Array3b()) is s.Array3b
    assert dr.mask_t(s.ArrayXf) is s.ArrayXb and dr.mask_t(s.ArrayXf()) is s.ArrayXb
    assert dr.mask_t(s.ArrayXb) is s.ArrayXb and dr.mask_t(s.ArrayXb()) is s.ArrayXb
    assert dr.mask_t(l.Float) is l.Bool and dr.mask_t(l.Float()) is l.Bool
    assert dr.mask_t(l.Bool) is l.Bool and dr.mask_t(l.Bool()) is l.Bool
    assert dr.mask_t(l.Array3f) is l.Array3b and dr.mask_t(l.Array3f()) is l.Array3b
    assert dr.mask_t(l.Array3b) is l.Array3b and dr.mask_t(l.Array3b()) is l.Array3b
    assert dr.mask_t(l.ArrayXf) is l.ArrayXb and dr.mask_t(l.ArrayXf()) is l.ArrayXb
    assert dr.mask_t(l.ArrayXb) is l.ArrayXb and dr.mask_t(l.ArrayXb()) is l.ArrayXb

    assert dr.is_integral_v(1) and dr.is_integral_v(int)
    assert dr.is_integral_v(s.Array3i()) and dr.is_integral_v(s.Array3i)
    assert not dr.is_integral_v(1.0) and not dr.is_integral_v(float)
    assert not dr.is_integral_v(s.Array3f()) and not dr.is_integral_v(s.Array3f)
    assert not dr.is_integral_v("str") and not dr.is_integral_v(str)
    assert not dr.is_integral_v(False) and not dr.is_integral_v(bool)
    assert not dr.is_integral_v(s.Array3b()) and not dr.is_integral_v(s.Array3b)

    assert not dr.is_float_v(1) and not dr.is_float_v(int)
    assert not dr.is_float_v(s.Array3i()) and not dr.is_float_v(s.Array3i)
    assert dr.is_float_v(1.0) and dr.is_float_v(float)
    assert dr.is_float_v(s.Array3f()) and dr.is_float_v(s.Array3f)
    assert not dr.is_float_v("str") and not dr.is_float_v(str)
    assert not dr.is_float_v(False) and not dr.is_float_v(bool)
    assert not dr.is_float_v(s.Array3b()) and not dr.is_float_v(s.Array3b)

    assert dr.is_arithmetic_v(1) and dr.is_arithmetic_v(int)
    assert dr.is_arithmetic_v(s.Array3i()) and dr.is_arithmetic_v(s.Array3i)
    assert dr.is_arithmetic_v(1.0) and dr.is_arithmetic_v(float)
    assert dr.is_arithmetic_v(s.Array3f()) and dr.is_arithmetic_v(s.Array3f)
    assert not dr.is_arithmetic_v("str") and not dr.is_arithmetic_v(str)
    assert not dr.is_arithmetic_v(False) and not dr.is_arithmetic_v(bool)
    assert not dr.is_arithmetic_v(s.Array3b()) and not dr.is_arithmetic_v(s.Array3b)

    assert not dr.is_mask_v(1) and not dr.is_mask_v(int)
    assert not dr.is_mask_v(s.Array3i()) and not dr.is_mask_v(s.Array3i)
    assert not dr.is_mask_v(1.0) and not dr.is_mask_v(float)
    assert not dr.is_mask_v(s.Array3f()) and not dr.is_mask_v(s.Array3f)
    assert not dr.is_mask_v("str") and not dr.is_mask_v(str)
    assert dr.is_mask_v(False) and dr.is_mask_v(bool)
    assert dr.is_mask_v(s.Array3b()) and dr.is_mask_v(s.Array3b)

    assert dr.uint32_array_t(float) is int
    assert dr.bool_array_t(float) is bool
    assert dr.float32_array_t(int) is float

    assert dr.bool_array_t(dr.scalar.Array3f) is dr.scalar.Array3b
    assert dr.int32_array_t(dr.scalar.Array3f) is dr.scalar.Array3i
    assert dr.uint32_array_t(dr.scalar.Array3f64) is dr.scalar.Array3u
    assert dr.int64_array_t(dr.scalar.Array3f) is dr.scalar.Array3i64
    assert dr.uint64_array_t(dr.scalar.Array3f) is dr.scalar.Array3u64
    assert dr.uint_array_t(dr.scalar.Array3f) is dr.scalar.Array3u
    assert dr.int_array_t(dr.scalar.Array3f) is dr.scalar.Array3i
    assert dr.uint_array_t(dr.scalar.Array3f64) is dr.scalar.Array3u64
    assert dr.int_array_t(dr.scalar.Array3f64) is dr.scalar.Array3i64
    assert dr.float_array_t(dr.scalar.Array3u) is dr.scalar.Array3f
    assert dr.float32_array_t(dr.scalar.Array3u) is dr.scalar.Array3f
    assert dr.float_array_t(dr.scalar.Array3u64) is dr.scalar.Array3f64
    assert dr.float32_array_t(dr.scalar.Array3u64) is dr.scalar.Array3f
    assert dr.float_array_t(dr.scalar.TensorXu64) is dr.scalar.TensorXf64


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test19_select():
    import drjit.scalar as s
    import drjit.llvm as l
    assert dr.select(True, "hello", "world") == "hello"
    result = dr.select(s.Array2b(True, False), 1, 2)
    assert isinstance(result, s.Array2i) and dr.all(result == s.Array2i(1, 2))
    result = dr.select(s.Array2b(True, False), 1, 2.0)
    assert isinstance(result, s.Array2f) and dr.all(result == s.Array2f(1, 2))
    result = dr.select(s.ArrayXb(True, False), 1, 2)
    assert isinstance(result, s.ArrayXi) and dr.all(result == s.ArrayXi(1, 2))
    result = dr.select(s.ArrayXb(True, False), 1, 2.0)
    assert isinstance(result, s.ArrayXf) and dr.all(result == s.ArrayXf(1, 2))

    result = dr.select(s.Array2b(True, False), s.Array2i(3, 4), s.Array2i(5, 6))
    assert isinstance(result, s.Array2i) and dr.all(result == s.Array2i(3, 6))
    result = dr.select(s.Array2b(True, False), s.Array2i(3, 4), s.Array2f(5, 6))
    assert isinstance(result, s.Array2f) and dr.all(result == s.Array2f(3, 6))
    result = dr.select(s.ArrayXb(True, False), s.ArrayXi(3, 4), s.ArrayXi(5, 6))
    assert isinstance(result, s.ArrayXi) and dr.all(result == s.ArrayXi(3, 6))
    result = dr.select(s.ArrayXb(True, False), s.ArrayXi(3, 4), s.ArrayXf(5, 6))
    assert isinstance(result, s.ArrayXf) and dr.all(result == s.ArrayXf(3, 6))

    result = dr.select(l.Array2b(True, False), 1, 2)
    assert isinstance(result, l.Array2i) and dr.all(result == l.Array2i(1, 2))
    result = dr.select(l.Array2b(True, False), 1, 2.0)
    assert isinstance(result, l.Array2f) and dr.all(result == l.Array2f(1, 2))
    result = dr.select(l.ArrayXb(True, False), 1, 2)
    assert isinstance(result, l.ArrayXi) and dr.all(result == l.ArrayXi(1, 2))
    result = dr.select(l.ArrayXb(True, False), 1, 2.0)
    assert isinstance(result, l.ArrayXf) and dr.all(result == l.ArrayXf(1, 2))

    result = dr.select(l.Array2b(True, False), l.Array2i(3, 4), l.Array2i(5, 6))
    assert isinstance(result, l.Array2i) and dr.all(result == l.Array2i(3, 6))
    result = dr.select(l.Array2b(True, False), l.Array2i(3, 4), l.Array2f(5, 6))
    assert isinstance(result, l.Array2f) and dr.all(result == l.Array2f(3, 6))
    result = dr.select(l.ArrayXb(True, False), l.ArrayXi(3, 4), l.ArrayXi(5, 6))
    assert isinstance(result, l.ArrayXi) and dr.all(result == l.ArrayXi(3, 6))
    result = dr.select(l.ArrayXb(True, False), l.ArrayXi(3, 4), l.ArrayXf(5, 6))
    assert isinstance(result, l.ArrayXf) and dr.all(result == l.ArrayXf(3, 6))


@pytest.mark.skip("nanobind layer")
def test20_component_access():
    from drjit.scalar import Array3f, Array4f
    from drjit.llvm import Array3f as Array3fL

    a = Array4f(4, 5, 6, 7)
    assert a.x == 4 and a.y == 5 and a.z == 6 and a.w == 7
    a.x, a.y, a.z, a.w = 1, 2, 3, 4
    assert a.x == 1 and a.y == 2 and a.z == 3 and a.w == 4
    a = Array3f(1, 2, 3)
    assert a.x == 1 and a.y == 2 and a.z == 3

    assert a.index == 0 and a.index_ad == 0

    with pytest.raises(TypeError) as ei:
        a.w == 4
    assert "Array3f: does not have a 'w' component!" in str(ei.value)

    a = Array3fL(1, 2, 3)
    assert a.index == 0 and a.index_ad == 0
    assert a.x.index != 0 and a.y.index != 0 and a.z.index != 0


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test21_zero_or_full():
    import drjit.scalar as s
    import drjit.llvm as l
    assert type(dr.zeros(dtype=int)) is int and dr.zeros(dtype=int) == 0
    assert type(dr.zeros(dtype=int, shape=(1,))) is int and dr.zeros(dtype=int, shape=(1,)) == 0
    assert type(dr.zeros(float)) is float and dr.zeros(float) == 0.0
    assert type(dr.zeros(float, shape=(1,))) is float and dr.zeros(float, shape=(1,)) == 0.0
    assert type(dr.zeros(float, shape=(100,))) is float and dr.zeros(float, shape=(100,)) == 0.0
    with pytest.raises(TypeError) as ei:
        dr.zeros(str)
    assert "Unsupported dtype!" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        dr.zeros(5)
    assert "incompatible function arguments" in str(ei.value)

    assert type(dr.zeros(s.Array3f)) is s.Array3f and dr.all(dr.zeros(s.Array3f) == s.Array3f(0))
    assert type(dr.zeros(s.Array3f, shape=(3,))) is s.Array3f and dr.all(dr.zeros(s.Array3f, shape=(3,)) == s.Array3f(0))
    assert type(dr.zeros(l.Array3f)) is l.Array3f and dr.all(dr.zeros(l.Array3f) == l.Array3f(0))
    assert type(dr.zeros(l.Array3f, shape=(3, 5))) is l.Array3f and dr.shape(dr.zeros(l.Array3f, shape=(3, 5))) == (3, 5)
    assert type(dr.zeros(l.Array3f, shape=10)) is l.Array3f and dr.shape(dr.zeros(l.Array3f, shape=10)) == (3, 10)
    assert type(dr.zeros(l.ArrayXf, shape=(8, 5))) is l.ArrayXf and dr.shape(dr.zeros(l.ArrayXf, shape=(8, 5))) == (8, 5)
    assert type(dr.zeros(l.Array3b, shape=10)) is l.Array3b and dr.shape(dr.zeros(l.Array3b, shape=10)) == (3, 10)
    assert type(dr.zeros(l.ArrayXb, shape=(8, 5))) is l.ArrayXb and dr.shape(dr.zeros(l.ArrayXb, shape=(8, 5))) == (8, 5)
    assert type(dr.full(l.ArrayXf, value=123, shape=(8, 5))) is l.ArrayXf and dr.all_nested(dr.full(l.ArrayXf, value=123, shape=(8, 5)) == 123)

    class Ray:
        DRJIT_STRUCT = { 'o': l.Array3f, 'd': l.Array3f, 'maxt' : l.Float, 'hit' : l.Bool }

    ray = dr.zeros(Ray, 10)
    assert isinstance(ray, Ray)
    assert isinstance(ray.o, l.Array3f) and isinstance(ray.d, l.Array3f) \
        and isinstance(ray.maxt, l.Float) and isinstance(ray.hit , l.Bool)

    assert ray.o.shape == (3, 10) and ray.d.shape == (3, 10) and ray.maxt.shape == (10,)
    assert dr.all_nested(ray.o == 0) and dr.all_nested(ray.d == 0) and dr.all(ray.maxt == 0) \
        and dr.all(ray.hit == False)

    ray = dr.ones(Ray, 10)
    assert dr.all_nested(ray.o == 1) and dr.all_nested(ray.d == 1) and dr.all(ray.maxt == 1) \
        and dr.all(ray.hit == True)


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test22_masked_assignment():
    import drjit.scalar as s
    import drjit.llvm as l

    a = s.Array3f(1,2,3)
    a[a <= 2] = 10
    assert dr.all(a == s.Array3f(10, 10, 3))

    a = l.Float(1,2,3)
    a[a <= 2] = 10
    assert dr.all(a == l.Float(10, 10, 3))

    a = l.Array3f(1,2,3)
    a[a <= 2] = 10
    assert dr.all(a == l.Array3f(10, 10, 3))


def test23_linspace():
    assert dr.allclose(2, 2)
    assert dr.allclose([1, 2, 3], [1, 2, 3])
    assert dr.allclose([1, 1, 1], 1)
    assert dr.allclose([[1, 1], [1, 1], [1, 1]], 1)
    assert not dr.allclose(2, 3)
    assert not dr.allclose([1, 2, 3], [1, 4, 3])
    assert not dr.allclose([1, 1, 1], 2)
    assert not dr.allclose(float('nan'), float('nan'))
    assert dr.allclose(float('nan'), float('nan'), equal_nan=True)

    with pytest.raises(RuntimeError) as ei:
        assert not dr.allclose([1,2,3], [1,4])
    assert 'incompatible sizes' in str(ei.value)

    np = pytest.importorskip("numpy")
    assert dr.allclose(np.array([1, 2, 3]), [1, 2, 3])
    assert dr.allclose(np.array([1, 2, 3]), dr.scalar.Array3f(1, 2, 3))
    assert dr.allclose(np.array([1, float('nan'), 3.0]), [1, float('nan'), 3], equal_nan=True)


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test24_arange():
    import drjit.scalar as s
    import drjit.llvm as l

    assert dr.all(dr.arange(s.ArrayXu, 5) == s.ArrayXu(0, 1, 2, 3, 4))
    assert dr.all(dr.arange(s.ArrayXf, 5) == s.ArrayXf(0, 1, 2, 3, 4))
    assert dr.all(dr.arange(s.ArrayXi, -2, 5, 2) == s.ArrayXi(-2, 0, 2, 4))
    assert dr.all(dr.arange(dtype=s.ArrayXf, start=-2, stop=5, step=2) == s.ArrayXf(-2, 0, 2, 4))

    assert dr.all(dr.arange(l.UInt, 5) == l.UInt(0, 1, 2, 3, 4))
    assert dr.all(dr.arange(l.Float, 5) == l.Float(0, 1, 2, 3, 4))
    assert dr.all(dr.arange(l.Int, -2, 5, 2) == l.Int(-2, 0, 2, 4))
    assert dr.all(dr.arange(dtype=l.Float, start=-2, stop=5, step=2) == l.Float(-2, 0, 2, 4))


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test25_linspace():
    import drjit.scalar as s
    import drjit.llvm as l

    assert dr.allclose(dr.linspace(s.ArrayXf, -2, 4, 4), s.ArrayXf(-2, 0, 2, 4))
    assert dr.allclose(dr.linspace(s.ArrayXf, start=-2, stop=5, num=4), s.ArrayXf(-2, 1/3, 8/3, 5))
    assert dr.allclose(dr.linspace(s.ArrayXf, start=-2, stop=4, num=4, endpoint=False), s.ArrayXf(-2, -0.5, 1, 2.5))
    assert dr.allclose(dr.linspace(l.Float, -2, 4, 4), l.Float(-2, 0, 2, 4))
    assert dr.allclose(dr.linspace(l.Float, start=-2, stop=5, num=4), l.Float(-2, 1/3, 8/3, 5))
    assert dr.allclose(dr.linspace(l.Float, start=-2, stop=4, num=4, endpoint=False), l.Float(-2, -0.5, 1, 2.5))


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test26_fast_cast(capsys):
    import drjit.llvm as l
    try:
        dr.set_log_level(5)
        x = l.Int(0, 1, 2)
        y = l.Float(x)
        z = x + 0.0

        a = l.Array3f(y, y, y)
        b = l.Array3u(a)

        out, err = capsys.readouterr()
        assert out.count('jit_var_new_cast') == 5
        assert out.count('jit_poke') == 3
    finally:
        dr.set_log_level(0)


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test27_gather_simple():
    import drjit.llvm as l
    import drjit.scalar as s

    assert dr.all(dr.gather(
        dtype=s.ArrayXf,
        source=dr.arange(s.ArrayXf, 10),
        index=s.ArrayXu(0, 5, 3)
    ) == s.ArrayXf(0, 5, 3))

    assert dr.all(dr.gather(
        dtype=s.ArrayXf,
        source=dr.arange(s.ArrayXf, 10),
        index=s.ArrayXu(0, 5, 3),
        active=s.ArrayXb(True, False, True)
    ) == s.ArrayXf(0, 0, 3))

    assert dr.all(dr.gather(
        dtype=l.Float,
        source=dr.arange(l.Float, 10),
        index=l.UInt(0, 5, 3)
    ) == l.Float(0, 5, 3))

    assert dr.all(dr.gather(
        dtype=l.Float,
        source=dr.arange(l.Float, 10),
        index=l.UInt(0, 5, 3),
        active=l.Bool(True, False, True)
    ) == l.Float(0, 0, 3))



@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test28_scatter_simple():
    import drjit.llvm as l
    import drjit.scalar as s

    target = dr.empty(s.ArrayXf, 3)
    dr.scatter(target, 1, [0, 1])
    dr.scatter(target, 2, 2)
    assert dr.all(target == s.ArrayXf([1, 1, 2]))

    target = dr.full(s.ArrayXf, shape=3, value=5)
    dr.scatter(target, [10, 12], [0, 1], [True, False])
    dr.scatter(target, 2, 2)
    assert dr.all(target == s.ArrayXf([10, 5, 2]))

    target = dr.empty(l.Float, 3)
    dr.scatter(target, 1, [0, 1])
    dr.scatter(target, 2, 2)
    assert dr.all(target == l.Float([1, 1, 2]))

    target = dr.full(l.Float, shape=3, value=5)
    dr.scatter(target, [10, 12], [0, 1], [True, False])
    dr.scatter(target, 2, 2)
    assert dr.all(target == l.Float([10, 5, 2]))


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test29_gather_complex():
    import drjit.scalar as s
    import drjit.llvm as l

    assert dr.all_nested(dr.gather(
        dtype=l.Array3f,
        source=dr.arange(l.Float, 12),
        index=l.UInt(0, 2, 3)
    ) == l.Array3f(
        l.Float(0, 6, 9),
        l.Float(1, 7, 10),
        l.Float(2, 8, 11),
    ))

    assert dr.all_nested(dr.gather(
        dtype=l.Array3f,
        source=dr.arange(l.Float, 12),
        index=l.UInt(0, 2, 3),
        active=l.Bool(True, False, True)
    ) == l.Array3f(
        l.Float(0, 0, 9),
        l.Float(1, 0, 10),
        l.Float(2, 0, 11),
    ))


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test30_gather_complex_2():
    x = dr.scalar.ArrayXf([1, 2, 3, 4])
    y = dr.scalar.ArrayXf([5, 6, 7, 8])
    i = dr.scalar.ArrayXu([1, 0])

    class MyStruct:
        DRJIT_STRUCT = { 'a' : dr.scalar.ArrayXf }
        def __init__(self, a: dr.scalar.ArrayXf=1):
            self.a = a

    x = MyStruct(x)
    r = dr.gather(MyStruct, x, i)
    assert type(r) is MyStruct
    assert dr.all(r.a == dr.scalar.ArrayXf([2, 1]))

    pytest.skip("nanobind layer")

    r = dr.gather(tuple, (x, y), i)
    assert type(r) is tuple and len(r) == 2
    assert dr.all(r[0] == dr.scalar.ArrayXf([2, 1]))
    assert dr.all(r[1] == dr.scalar.ArrayXf([6, 5]))


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test31_scatter_complex():
    import drjit.scalar as s
    import drjit.llvm as l

    target = dr.empty(l.Float, 6)
    # dr.scatter(target, l.Array3f([[1, 4], [2, 5], [3, 6]]), [0, 1]) # TODO
    dr.scatter(target, l.Array3f([[1, 4], [2, 5], [3, 6]]), l.UInt([0, 1]))

    assert dr.all(target == l.Float([1, 2, 3, 4, 5, 6]))


@pytest.mark.skip("nanobind layer")
def test31_scatter_complex_2():
    x = dr.scalar.ArrayXf([1, 2, 3, 4])
    y = dr.scalar.ArrayXf([5, 6, 7, 8])
    a = dr.scalar.ArrayXf([10, 11])
    b = dr.scalar.ArrayXf([20, 21])
    i = dr.scalar.ArrayXu([1, 0])
    dr.scatter((x, y), (a, b), i)
    assert dr.all(x == dr.scalar.ArrayXf([11, 10, 3, 4]))
    assert dr.all(y == dr.scalar.ArrayXf([21, 20, 7, 8]))

    class MyStruct:
        DRJIT_STRUCT = { 'a' : dr.scalar.ArrayXf }
        def __init__(self, a: dr.scalar.ArrayXf):
            self.a = a

    x = MyStruct(x)
    y = MyStruct(b)
    dr.scatter(x, y, i)
    assert dr.all(x.a == dr.scalar.ArrayXf(21, 20, 3, 4))


@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test32_ravel(capsys):
    import drjit.scalar as s
    import drjit.llvm as l

    try:
        dr.set_log_level(5)

        assert dr.all(dr.ravel(s.Array3f(1, 2, 3), order='C') == s.ArrayXf([1, 2, 3]))
        assert dr.all(dr.ravel(s.Array3f(1, 2, 3), order='F') == s.ArrayXf([1, 2, 3]))

        x = l.Array3f([1, 2], [3, 4], [5, 6])
        assert dr.all(dr.ravel(x.x) is x.x)
        assert dr.all(dr.ravel(x, order='C') == l.Float([1, 2, 3, 4, 5, 6]))
        assert dr.all(dr.ravel(x) == l.Float([1, 3, 5, 2, 4, 6]))

        out, err = capsys.readouterr()
        assert out.count('jit_var_new_scatter') == 6
        assert out.count('jit_poke') == 18
    finally:
        dr.set_log_level(0)


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test33_unravel(capsys):
    import drjit.scalar as s
    import drjit.llvm as l

    try:
        dr.set_log_level(5)

        assert dr.all(dr.unravel(s.Array3f, s.ArrayXf([1, 2, 3]), order='C') == s.Array3f(1, 2, 3))
        assert dr.all(dr.unravel(s.Array3f, s.ArrayXf([1, 2, 3]), order='F') == s.Array3f(1, 2, 3))

        x = l.Float([1, 2, 3, 4, 5, 6])
        assert dr.all(dr.unravel(l.Float, x) is x)
        assert dr.all_nested(dr.unravel(l.Array3f, x, order='C') == l.Array3f([1, 2], [3, 4], [5, 6]))
        assert dr.all_nested(dr.unravel(l.Array3f, x) == l.Array3f([1, 4], [2, 5], [3, 6]))

        out, err = capsys.readouterr()
        assert out.count('jit_var_new_gather') == 6
        assert out.count('jit_poke') == 18
    finally:
        dr.set_log_level(0)


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test34_schedule(capsys):
    import drjit.llvm as l
    try:
        class MyStruct:
            DRJIT_STRUCT = { 'a' : l.Float }
            def __init__(self, a: l.Float):
                self.a = a

        dr.set_log_level(5)
        assert dr.schedule() is False
        assert dr.schedule(123, [], [123], (), (123,), {123: 456}) is False
        x = l.Float([1, 2]) + 3
        assert dr.schedule(x) is True

        y = l.Array3f(x + 1, x + 2, x + 3)
        assert dr.schedule({'hello': [(y, 4)]}) is True
        z = MyStruct(x + 4)
        dr.schedule(z)
        out, err = capsys.readouterr()
        assert out.count('jit_var_schedule') == 5
    finally:
        dr.set_log_level(0)


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test35_to_dlpack_numpy_cpu():
    import drjit.scalar as s
    import drjit.llvm as l

    from_dlpack = None
    try:
        import numpy as np
        from_dlpack = getattr(np, '_from_dlpack', None)
        from_dlpack = getattr(np, 'from_dlpack', from_dlpack)
    except:
        pass

    if from_dlpack is None:
        pytest.skip('NumPy is missing/too old')

    x = l.Array3f([1, 2], [3, 4], [5, 6])
    assert x.__dlpack_device__() == (1, 0)
    assert np.all(from_dlpack(x) == np.array([[1, 2], [3, 4], [5, 6]]))
    assert np.all(x.__array__() == np.array([[1, 2], [3, 4], [5, 6]]))

    x = s.ArrayXf(1, 2, 3, 4)
    assert np.all(x.__array__() == np.array([1, 2, 3, 4]))
    x = s.Array3f(1, 2, 3)
    assert np.all(x.__array__() == np.array([1, 2, 3]))


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.CUDA), reason='CUDA mode is unsupported')
def test36_to_dlpack_numpy_gpu():
    np = pytest.importorskip("numpy")

    import drjit.cuda as c

    x = c.Array3f([1, 2], [3, 4], [5, 6])
    assert np.all(x.__array__() == np.array([[1, 2], [3, 4], [5, 6]]))


def test38_construct_from_numpy_1():
    # Simple scalar conversions, different types, static sizes
    np = pytest.importorskip("numpy")

    import drjit.scalar as s

    assert dr.all(s.Array3f(np.array([1, 2, 3], dtype=np.float32)) == s.Array3f(1, 2, 3))
    assert dr.all(s.Array3f(np.array([1, 2, 3], dtype=np.float64)) == s.Array3f(1, 2, 3))
    assert dr.all(s.Array3f(np.array([1, 2, 3], dtype=np.int32)) == s.Array3f(1, 2, 3))

    with pytest.raises(TypeError):
        # Size mismatch
        s.Array3f(np.array([1, 2, 3, 4], dtype=np.float32))


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test39_construct_from_numpy_2():
    # Dynamically allocated arrays + implicit casts
    np = pytest.importorskip("numpy")

    import drjit.scalar as s
    import drjit.llvm as l

    p = np.array([1, 2, 3, 4], dtype=np.float32)
    r = s.ArrayXf(p)
    assert dr.all(r == s.ArrayXf(1, 2, 3, 4))

    p = np.array([1, 2, 3, 4], dtype=np.float32)
    r = l.Float(p)
    assert dr.all(r == l.Float(1, 2, 3, 4))

    # Check if zero-copy constructor works
    p[0] = 5
    assert dr.all(r == l.Float(5, 2, 3, 4))

    r = l.Float64(p)
    assert dr.all(r == l.Float64(5, 2, 3, 4))
    p[0] = 1
    assert dr.all(r == l.Float64(5, 2, 3, 4))

    with pytest.raises(TypeError) as ei:
        l.Array4f(p)
    assert "unable to initialize from tensor of type 'numpy.ndarray'. The input must have the following configuration for this to succeed: shape=(4, *), dtype=float32, order='C'" in str(ei.value)


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test40_construct_from_numpy_3():
    # Nested arrays, CPU-only
    np = pytest.importorskip("numpy")

    import drjit.llvm as l

    p = np.array([[1, 2], [4, 5], [6, 7]], dtype=np.float32)

    with pytest.raises(TypeError) as ei:
        r = l.Float(p)
    assert "The input must have the following configuration for this to succeed: shape=(*), dtype=float32, order='C'" in str(ei.value)

    r = l.Array3f(p)
    assert dr.all_nested(r == l.Array3f([1, 2], [4, 5], [6, 7]))


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.CUDA), reason='CUDA mode is unsupported')
def test41_construct_from_numpy_4():
    # Nested arrays, copy to CUDA Dr.Jit array
    np = pytest.importorskip("numpy")

    import drjit.cuda as c

    p = np.array([[1, 2], [4, 5], [6, 7]], dtype=np.float32)

    r = c.Array3f(p)
    assert dr.all_nested(r == c.Array3f([1, 2], [4, 5], [6, 7]))
    r = c.Array3i64(p)
    assert dr.all_nested(r == c.Array3i64([1, 2], [4, 5], [6, 7]))


@pytest.mark.skip("nanobind layer")
@pytest.mark.skipif(not dr.has_backend(dr.JitBackend.LLVM), reason='LLVM mode is unsupported')
def test42_prevent_inefficient_cast(capsys):
    import drjit.scalar as s
    import drjit.llvm as l
    import drjit.llvm.ad as la

    with pytest.raises(TypeError) as ei:
        s.ArrayXf(la.Float([1, 2, 3, 4]))

    with pytest.raises(TypeError) as ei:
        l.Float(s.ArrayXf([1, 2, 3, 4]))

    with pytest.raises(TypeError) as ei:
        l.Float(la.Float([1, 2, 3, 4]))

    with pytest.warns(RuntimeWarning, match=r"implicit conversion"):
        with pytest.raises(TypeError) as ei:
            l.Array3f(la.Array3f(1))

    #  la.Float(l.Float([1, 2, 3, 4]))


@pytest.fixture(scope="module", params=['drjit.cuda', 'drjit.llvm'])
def m(request):
    if 'cuda' in request.param:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    else:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')
    yield importlib.import_module(request.param)


def test43_minimum(m):
    assert dr.allclose(dr.minimum(6.0, 4.0), 4.0)

    a = dr.minimum(m.Float([1, 2, 3]), m.Float(2))
    assert dr.allclose(a, [1, 2, 2])
    assert type(a) is m.Float

    a = dr.minimum(m.Float([1, 2, 3]), [2.0, 2.0, 2.0])
    assert dr.allclose(a, [1, 2, 2])
    assert type(a) is m.Float

    a = dr.minimum(m.Array3f(1, 2, 3), m.Float(2))
    assert dr.allclose(a, [1, 2, 2])
    assert type(a) is m.Array3f

    a = dr.minimum(m.Array3i(1, 2, 3), m.Float(2))
    assert dr.allclose(a, [1, 2, 2])
    assert type(a) is m.Array3f

    a = dr.minimum(m.ArrayXf([1, 2, 5], [2, 1, 4], [3, 4, 3]), m.Float(1, 2, 3))
    assert dr.allclose(a, [[1, 2, 3], [1, 1, 3], [1, 2, 3]])
    assert type(a) is m.ArrayXf

    a = dr.minimum(m.ArrayXf([1, 2, 5], [2, 1, 4], [3, 4, 3]), m.Array3f(1, 2, 3))
    assert dr.allclose(a, [[1, 1, 1], [2, 1, 2], [3, 3, 3]])
    assert type(a) is m.ArrayXf


def test44_maximum(m):
    assert dr.allclose(dr.maximum(6.0, 4.0), 6.0)

    a = dr.maximum(m.Float([1, 2, 3]), m.Float(2))
    assert dr.allclose(a, [2, 2, 3])
    assert type(a) is m.Float

    a = dr.maximum(m.Float([1, 2, 3]), [2.0, 2.0, 2.0])
    assert dr.allclose(a, [2, 2, 3])
    assert type(a) is m.Float

    a = dr.maximum(m.Array3f(1, 2, 3), m.Float(2))
    assert dr.allclose(a, [2, 2, 3])
    assert type(a) is m.Array3f

    a = dr.maximum(m.Array3i(1, 2, 3), m.Float(2))
    assert dr.allclose(a, [2, 2, 3])
    assert type(a) is m.Array3f

    a = dr.maximum(m.ArrayXf([1, 2, 5], [2, 1, 4], [3, 4, 3]), m.Float(1, 2, 3))
    assert dr.allclose(a, [[1, 2, 5], [2, 2, 4], [3, 4, 3]])
    assert type(a) is m.ArrayXf

    a = dr.maximum(m.ArrayXf([1, 2, 5], [2, 1, 4], [3, 4, 3]), m.Array3f(1, 2, 3))
    assert dr.allclose(a, [[1, 2, 5], [2, 2, 4], [3, 4, 3]])
    assert type(a) is m.ArrayXf


def test45_iter(m):
    a = m.Float([0, 1, 2, 3, 4 ,5])
    for i, v in enumerate(a):
        assert v == i

    a = m.Array0f()
    count = 0
    for i, v in enumerate(a):
        count += 1
    assert count == 0

    a = m.Array1f([0, 1, 2, 3, 4 ,5])
    count = 0
    for i, v in enumerate(a):
        assert dr.allclose(v, [0, 1, 2, 3, 4 ,5])
        count += 1
    assert count == 1

    a = m.Array3f([[0, 0], [1, 1], [2, 2]])
    count = 0
    for i, v in enumerate(a):
        assert dr.allclose(v, i)
        count += 1
    assert count == 3


#@pytest.mark.parametrize('name', ['sqrt', 'cbrt', 'sin', 'cos', 'tan', 'asin',
#                                  'acos', 'atan', 'sinh', 'cosh', 'tanh',
#                                  'asinh', 'acosh', 'atanh', 'exp', 'exp2',
#                                  'log', 'log2', 'floor', 'ceil', 'trunc',
#                                  'round', 'rcp', 'rsqrt'])
#def test17_spotcheck_unary_math(name):
#    from drjit.scalar import ArrayXf, PCG32
#    import math
#    func_ref = getattr(math, name, None)
#    if name == 'cbrt':
#        func_ref = lambda x: x**(1/3)
#    elif name == 'exp2':
#        func_ref = lambda x: 2**x
#    elif name == 'log2':
#        log2 = lambda x: math.log(x) / math.log(2)
#    elif name == 'round':
#        func_ref = round
#    elif name == 'rcp':
#        func_ref = lambda x : 1/x
#    elif name == 'rsqrt':
#        func_ref = lambda x : math.sqrt(1/x)
#
#    rng = PCG32()
#    x = ArrayXf((rng.next_float32() for i in range(10)))
#    if name == 'acosh':
#        x += 1
#    ref = ArrayXf([func_ref(y) for y in x])
#    func = getattr(dr, name)
#    value = func(x)
#    value_2 = ArrayXf(func(y) for y in x)
#    assert dr.allclose(value, func)
#




from drjit import Dynamic, Exception, VarType
import drjit as _dr
import weakref as _wr


def _check1(a0):
    s0 = len(a0)
    ar = a0.empty_(s0 if a0.Size == Dynamic else 0)
    return (ar, s0)


def _check2(a0, a1):
    """Validate the inputs of a binary generic array operation"""
    s0, s1 = len(a0), len(a1)
    sr = max(s0, s1)
    if (s0 != sr and s0 != 1) or (s1 != sr and s1 != 1):
        raise Exception("Incompatible argument sizes: %i and %i" % (s0, s1))
    elif type(a0) is not type(a1):  # noqa
        raise Exception("Type mismatch!")
    ar = a0.empty_(sr if a0.Size == Dynamic else 0)
    return (ar, sr)


def _check2_inplace(a0, a1):
    """Validate the inputs of a binary generic array operation (in-place)"""
    s0, s1 = len(a0), len(a1)
    sr = max(s0, s1)
    if s0 != sr or (s1 != sr and s1 != 1):
        raise Exception("Incompatible argument sizes: %i and %i" % (s0, s1))
    elif type(a0) is not type(a1):  # noqa
        raise Exception("Type mismatch!")
    return sr


def _check2_bitop(a0, a1):
    """Validate the inputs of a binary bit manipulation array operation"""
    s0, s1 = len(a0), len(a1)
    sr = max(s0, s1)
    if (s0 != sr and s0 != 1) or (s1 != sr and s1 != 1):
        raise Exception("Incompatible argument sizes: %i and %i" % (s0, s1))
    elif type(a0) is not type(a1) and type(a1) is not a0.MaskType:  # noqa
        raise Exception("Type mismatch!")
    ar = a0.empty_(sr if a0.Size == Dynamic else 0)
    return (ar, sr)


def _check2_bitop_inplace(a0, a1):
    """Validate the inputs of a binary bit manipulation array operation
       (in-place)"""
    s0, s1 = len(a0), len(a1)
    sr = max(s0, s1)
    if s0 != sr or (s1 != sr and s1 != 1):
        raise Exception("Incompatible argument sizes: %i and %i" % (s0, s1))
    elif type(a0) is not type(a1) and type(a1) is not a0.MaskType:  # noqa
        raise Exception("Type mismatch!")
    return sr


def _check2_mask(a0, a1):
    """Validate the inputs of a binary mask-producing array operation"""
    s0, s1 = len(a0), len(a1)
    sr = max(s0, s1)
    if (s0 != sr and s0 != 1) or (s1 != sr and s1 != 1):
        raise Exception("Incompatible argument sizes: %i and %i" % (s0, s1))
    elif type(a0) is not type(a1):  # noqa
        raise Exception("Type mismatch!")
    ar = a0.MaskType.empty_(sr if a0.Size == Dynamic else 0)
    return (ar, sr)


def _check3(a0, a1, a2):
    """Validate the inputs of a ternary generic array operation"""
    s0, s1, s2 = len(a0), len(a1), len(a2)
    sr = max(s0, s1, s2)
    if (s0 != sr and s0 != 1) or (s1 != sr and s1 != 1) or \
       (s2 != sr and s2 != 1):
        raise Exception("Incompatible argument sizes: %i, %i, and %i"
                        % (s0, s1, s2))
    elif type(a0) is not type(a1) or type(a2) is not type(a2):  # noqa
        raise Exception("Type mismatch!")
    ar = a0.empty_(sr if a0.Size == Dynamic else 0)
    return (ar, sr)


def _check3_select(a0, a1, a2):
    """Validate the inputs of a select() array operation"""
    s0, s1, s2 = len(a0), len(a1), len(a2)
    sr = max(s0, s1, s2)
    if (s0 != sr and s0 != 1) or (s1 != sr and s1 != 1) or \
       (s2 != sr and s2 != 1):
        raise Exception("Incompatible argument sizes: %i, %i, and %i"
                        % (s0, s1, s2))
    elif type(a1) is not type(a2) or type(a0) is not type(a1).MaskType:  # noqa
        raise Exception("Type mismatch!")
    ar = a1.empty_(sr if a0.Size == Dynamic else 0)
    return (ar, sr)


def _binary_op(a, b, fn):
    """
    Perform a bit-level operation 'fn' involving variables 'a' and 'b' in a way
    that works even when the operands are floating point variables
    """

    convert = isinstance(a, float) or isinstance(b, float)

    if convert:
        src, dst = VarType.Float64, VarType.Int64
        a = _dr.detail.reinterpret_scalar(a, src, dst)
        if isinstance(b, bool):
            b = -1 if b else 0
        else:
            b = _dr.detail.reinterpret_scalar(b, src, dst)

    c = fn(a, b)

    if convert:
        c = _dr.detail.reinterpret_scalar(c, dst, src)

    return c

# -------------------------------------------------------------------
#                        Vertical operations
# -------------------------------------------------------------------

def neg_(a0):
    if not a0.IsArithmetic:
        raise Exception("neg(): requires arithmetic operands!")
    ar, sr = _check1(a0)
    for i in range(sr):
        ar[i] = -a0[i]
    return ar


def not_(a0):
    if a0.IsFloat:
        raise Exception("not(): requires an integral or mask operand!")
    ar, sr = _check1(a0)
    if a0.Value is bool:
        for i in range(sr):
            ar[i] = not a0[i]
    else:
        for i in range(sr):
            ar[i] = ~a0[i]
    return ar


def add_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("add(): requires arithmetic operands!")
    ar, sr = _check2(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] + a1[i]
    return ar


def iadd_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("iadd(): requires arithmetic operands!")
    sr = _check2_inplace(a0, a1)
    for i in range(sr):
        a0[i] += a1[i]
    return a0


def sub_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("sub(): requires arithmetic operands!")
    ar, sr = _check2(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] - a1[i]
    return ar


def isub_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("isub(): requires arithmetic operands!")
    sr = _check2_inplace(a0, a1)
    for i in range(sr):
        a0[i] -= a1[i]
    return a0


def mul_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("mul(): requires arithmetic operands!")

    if _dr.depth_v(a1) < _dr.depth_v(a0):
        # Avoid type promotion in scalars multiplication, which would
        # be costly for special types (matrices, quaternions, etc.)
        sr = len(a0)
        ar = a0.empty_(sr if a0.Size == Dynamic else 0)
        for i in range(len(a0)):
            ar[i] = a0[i] * a1
        return ar

    ar, sr = _check2(a0, a1)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = a0[i] * a1[i]
    elif a0.IsComplex:
        ar.real = _dr.fma(a0.real, a1.real, -a0.imag * a1.imag)
        ar.imag = _dr.fma(a0.real, a1.imag, a0.imag * a1.real)
    elif a0.IsQuaternion:
        tbl = (4, 3, -2, 1, -3, 4, 1, 2, 2, -1, 4, 3, -1, -2, -3, 4)
        for i in range(4):
            accum = 0
            for j in range(4):
                idx = tbl[i*4 + j]
                value = a1[abs(idx) - 1]
                accum = _dr.fma(a0[j], value if idx > 0 else -value, accum)
            ar[i] = accum
    elif a0.IsMatrix:
        raise Exception("mul(): please use the matrix multiplication operator '@' instead.")
    else:
        raise Exception("mul(): unsupported array type!")
    return ar


def imul_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("imul(): requires arithmetic operands!")
    sr = _check2_inplace(a0, a1)
    if not a0.IsSpecial:
        for i in range(sr):
            a0[i] *= a1[i]
    else:
        a0.assign(a0 * a1)
    return a0


def matmul_(a0, a1):
    if not (a0.Size == a1.Size and (a0.IsMatrix or a0.IsVector) \
                               and (a1.IsMatrix or a1.IsVector)):
        raise Exception("matmul(): unsupported operand shape!")

    if a0.IsVector and a1.IsVector:
        return _dr.dot(a0, a1)
    elif a0.IsMatrix and a1.IsVector:
        ar = a0[0] * a1[0]
        for i in range(1, a1.Size):
            ar = _dr.fma(a0[i], a1[i], ar)
        return ar
    elif a0.IsVector and a1.IsMatrix:
        ar = a1.Value()
        for i in range(a1.Size):
            ar[i] = _dr.dot(a0, a1[i])
        return ar
    else: # matrix @ matrix
        ar, sr = _check2(a0, a1)
        for j in range(a0.Size):
            accum = a0[0] * _dr.full(a0.Value, a1[0, j])
            for i in range(1, a0.Size):
                accum = _dr.fma(a0[i], _dr.full(a0.Value, a1[i, j]), accum)
            ar[j] = accum
        return ar

def truediv_(a0, a1):
    if not a0.IsFloat:
        raise Exception("Use the floor division operator \"//\" for "
                        "Dr.Jit integer arrays.")
    if not a0.IsSpecial:
        ar, sr = _check2(a0, a1)
        for i in range(sr):
            ar[i] = a0[i] / a1[i]
        return ar
    elif a0.IsSpecial:
        return a0 * a1.rcp_()
    else:
        raise Exception("truediv(): unsupported array type!")


def itruediv_(a0, a1):
    if not a0.IsFloat:
        raise Exception("Use the floor division operator \"//=\" for "
                        "Dr.Jit integer arrays.")
    sr = _check2_inplace(a0, a1)
    if not a0.IsSpecial:
        for i in range(sr):
            a0[i] /= a1[i]
    else:
        a0.assign(a0 / a1)

    return a0


def floordiv_(a0, a1):
    if not a0.IsIntegral:
        raise Exception("Use the true division operator \"/\" for "
                        "Dr.Jit floating point arrays.")
    ar, sr = _check2(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] // a1[i]
    return ar


def ifloordiv_(a0, a1):
    if not a0.IsIntegral:
        raise Exception("Use the floor division operator \"/=\" for "
                        "Dr.Jit floating point arrays.")
    sr = _check2_inplace(a0, a1)
    for i in range(sr):
        a0[i] //= a1[i]
    return a0


def mod_(a0, a1):
    if not a0.IsIntegral:
        raise Exception("mod(): requires arithmetic operands!")
    ar, sr = _check2(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] % a1[i]
    return ar


def imod_(a0, a1):
    if not a0.IsIntegral:
        raise Exception("imod(): requires arithmetic operands!")
    sr = _check2_inplace(a0, a1)
    for i in range(sr):
        a0[i] %= a1[i]
    return a0


def and_(a0, a1):
    ar, sr = _check2_bitop(a0, a1)

    if ar.Depth > 1:
        for i in range(sr):
            ar[i] = a0[i] & a1[i]
    else:
        for i in range(sr):
            ar[i] = _binary_op(a0[i], a1[i],
                               lambda a, b: a & b)

    return ar


def iand_(a0, a1):
    sr = _check2_bitop_inplace(a0, a1)
    if a0.Depth > 1:
        for i in range(sr):
            a0[i] &= a1[i]
    else:
        for i in range(sr):
            a0[i] = _binary_op(a0[i], a1[i],
                               lambda a, b: a & b)
    return a0


def or_(a0, a1):
    ar, sr = _check2_bitop(a0, a1)

    if ar.Depth > 1:
        for i in range(sr):
            ar[i] = a0[i] | a1[i]
    else:
        for i in range(sr):
            ar[i] = _binary_op(a0[i], a1[i],
                               lambda a, b: a | b)

    return ar


def ior_(a0, a1):
    sr = _check2_bitop_inplace(a0, a1)
    if a0.Depth > 1:
        for i in range(sr):
            a0[i] |= a1[i]
    else:
        for i in range(sr):
            a0[i] = _binary_op(a0[i], a1[i],
                               lambda a, b: a | b)
    return a0


def xor_(a0, a1):
    ar, sr = _check2_bitop(a0, a1)

    if ar.Depth > 1:
        for i in range(sr):
            ar[i] = a0[i] ^ a1[i]
    else:
        for i in range(sr):
            ar[i] = _binary_op(a0[i], a1[i],
                               lambda a, b: a ^ b)

    return ar


def ixor_(a0, a1):
    sr = _check2_bitop_inplace(a0, a1)
    if a0.Depth > 1:
        for i in range(sr):
            a0[i] ^= a1[i]
    else:
        for i in range(sr):
            a0[i] = _binary_op(a0[i], a1[i],
                               lambda a, b: a ^ b)
    return a0


def sl_(a0, a1):
    if not a0.IsIntegral:
        raise Exception("sl(): requires integral operands!")
    ar, sr = _check2(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] << a1[i]
    return ar


def isl_(a0, a1):
    if not a0.IsIntegral:
        raise Exception("isl(): requires integral operands!")
    sr = _check2_inplace(a0, a1)
    for i in range(sr):
        a0[i] <<= a1[i]
    return a0


def sr_(a0, a1):
    if not a0.IsIntegral:
        raise Exception("sr(): requires integral operands!")
    ar, sr = _check2(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] >> a1[i]
    return ar


def isr_(a0, a1):
    if not a0.IsIntegral:
        raise Exception("isr(): requires integral operands!")
    sr = _check2_inplace(a0, a1)
    for i in range(sr):
        a0[i] >>= a1[i]
    return a0


def lt_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("lt(): requires arithmetic operands!")
    ar, sr = _check2_mask(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] < a1[i]
    return ar


def le_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("le(): requires arithmetic operands!")
    ar, sr = _check2_mask(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] <= a1[i]
    return ar


def gt_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("gt(): requires arithmetic operands!")
    ar, sr = _check2_mask(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] > a1[i]
    return ar


def ge_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("ge(): requires arithmetic operands!")
    ar, sr = _check2_mask(a0, a1)
    for i in range(sr):
        ar[i] = a0[i] >= a1[i]
    return ar


def eq_(a0, a1):
    ar, sr = _check2_mask(a0, a1)
    for i in range(sr):
        ar[i] = _dr.eq(a0[i], a1[i])
    return ar


def neq_(a0, a1):
    ar, sr = _check2_mask(a0, a1)
    for i in range(sr):
        ar[i] = _dr.neq(a0[i], a1[i])
    return ar


def sqrt_(a0):
    if not a0.IsFloat:
        raise Exception("sqrt(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.sqrt(a0[i])
    elif a0.IsComplex:
        n = abs(a0)
        m = a0.real >= 0
        zero = _dr.eq(n, 0)
        t1 = _dr.sqrt(.5 * (n + abs(a0.real)))
        t2 = .5 * a0.imag / t1
        im = _dr.select(m, t2, _dr.copysign(t1, a0.imag))
        ar.real = _dr.select(m, t1, abs(t2))
        ar.imag = _dr.select(zero, 0, im)
    elif a0.IsQuaternion:
        ri = _dr.norm(a0.imag)
        cs = _dr.sqrt(a0.Complex(a0.real, ri))
        ar.imag = a0.imag * (_dr.rcp(ri) * cs.imag)
        ar.real = cs.real
    else:
        raise Exception("sqrt(): unsupported array type!")
    return ar


def rsqrt_(a0):
    if not a0.IsFloat:
        raise Exception("rsqrt(): requires floating point operands!")
    if not a0.IsSpecial:
        ar, sr = _check1(a0)
        for i in range(sr):
            ar[i] = _dr.rsqrt(a0[i])
        return ar
    else:
        return _dr.rcp(_dr.sqrt(a0))


def rcp_(a0):
    if not a0.IsFloat:
        raise Exception("rcp(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.rcp(a0[i])
    elif a0.IsComplex or a0.IsQuaternion:
        return _dr.conj(a0) * _dr.rcp(_dr.squared_norm(a0))
    else:
        raise Exception("rcp(): unsupported array type!")
    return ar


def abs_(a0):
    if not a0.IsArithmetic:
        raise Exception("abs(): requires arithmetic operands!")
    if not a0.IsSpecial or a0.IsMatrix:
        ar, sr = _check1(a0)
        for i in range(sr):
            ar[i] = _dr.abs(a0[i])
        return ar
    elif a0.IsSpecial:
        return _dr.norm(a0)
    else:
        raise Exception("abs(): unsupported array type!")


def mulhi_(a0, a1):
    ar, sr = _check2(a0, a1)
    if not a0.IsIntegral:
        raise Exception("mulhi(): requires integral operands!")
    for i in range(sr):
        ar[i] = _dr.mulhi(a0[i], a1[i])
    return ar


def maximum_(a0, a1):
    ar, sr = _check2(a0, a1)
    if not a0.IsArithmetic:
        raise Exception("maximum(): requires arithmetic operands!")
    for i in range(sr):
        ar[i] = _dr.maximum(a0[i], a1[i])
    return ar


def minimum_(a0, a1):
    if not a0.IsArithmetic:
        raise Exception("minimum(): requires arithmetic operands!")
    ar, sr = _check2(a0, a1)
    for i in range(sr):
        ar[i] = _dr.minimum(a0[i], a1[i])
    return ar


def fma_(a0, a1, a2):
    if not a0.IsFloat:
        raise Exception("fma(): requires floating point operands!")
    if not a0.IsSpecial:
        ar, sr = _check3(a0, a1, a2)
        for i in range(sr):
            ar[i] = _dr.fma(a0[i], a1[i], a2[i])
        return ar
    else:
        return a0 * a1 + a2


def tzcnt_(a0):
    if not a0.IsIntegral:
        raise Exception("tzcnt(): requires integral operands!")
    ar, sr = _check1(a0)
    for i in range(sr):
        ar[i] = _dr.tzcnt(a0[i])
    return ar


def lzcnt_(a0):
    if not a0.IsIntegral:
        raise Exception("lzcnt(): requires integral operands!")
    ar, sr = _check1(a0)
    for i in range(sr):
        ar[i] = _dr.lzcnt(a0[i])
    return ar


def popcnt_(a0):
    if not a0.IsIntegral:
        raise Exception("popcnt(): requires integral operands!")
    ar, sr = _check1(a0)
    for i in range(sr):
        ar[i] = _dr.popcnt(a0[i])
    return ar


def floor_(a0):
    if not a0.IsArithmetic:
        raise Exception("floor(): requires arithmetic operands!")
    ar, sr = _check1(a0)
    for i in range(sr):
        ar[i] = _dr.floor(a0[i])
    return ar


def ceil_(a0):
    if not a0.IsArithmetic:
        raise Exception("ceil(): requires arithmetic operands!")
    ar, sr = _check1(a0)
    for i in range(sr):
        ar[i] = _dr.ceil(a0[i])
    return ar


def round_(a0):
    if not a0.IsArithmetic:
        raise Exception("round(): requires arithmetic operands!")
    ar, sr = _check1(a0)
    for i in range(sr):
        ar[i] = _dr.round(a0[i])
    return ar


def trunc_(a0):
    if not a0.IsArithmetic:
        raise Exception("trunc(): requires arithmetic operands!")
    ar, sr = _check1(a0)
    for i in range(sr):
        ar[i] = _dr.trunc(a0[i])
    return ar


@staticmethod
def select_(a0, a1, a2):
    ar, sr = _check3_select(a0, a1, a2)
    for i in range(sr):
        ar[i] = _dr.select(a0[i], a1[i], a2[i])
    return ar

# -------------------------------------------------------------------
#       Vertical operations -- ad/JIT compilation-related
# -------------------------------------------------------------------


def label_(a):
    if a.IsTensor:
        return a.array.label_()
    if a.IsJIT or a.IsDiff:
        return [v.label_() for v in a]
    return None


def set_label_(a, label):
    if a.IsTensor:
        return a.array.set_label_(label)
    if a.IsJIT or a.IsDiff:
        if isinstance(label, tuple) or isinstance(label, list):
            if len(a) != len(label):
                raise Exception("Size mismatch!")
            for i, v in enumerate(a):
                v.set_label_(label[i])
        else:
            for i in range(len(a)):
                a[i].set_label_(label + "_%i" % i)


@property
def label(a):
    return a.label_()


@label.setter
def label(a, value):
    return a.set_label_(value)


def schedule_(a0):
    result = False
    if a0.IsTensor:
        result |= a0.array.schedule_()
    elif a0.IsJIT:
        if a0.Depth == 1:
            if a0.IsDiff:
                a0 = a0.detach_()
            result |= _dr.detail.schedule(a0.index)
        else:
            for i in range(len(a0)):
                result |= a0[i].schedule_()
    return result


# -------------------------------------------------------------------
#           Vertical operations -- transcendental functions
# -------------------------------------------------------------------


def sin_(a0):
    if not a0.IsFloat:
        raise Exception("sin(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.sin(a0[i])
    elif a0.IsComplex:
        s, c = _dr.sincos(a0.real)
        sh, ch = _dr.sincosh(a0.imag)
        ar.real = s * ch
        ar.imag = c * sh
    else:
        raise Exception("sin(): unsupported array type!")
    return ar


def cos_(a0):
    if not a0.IsFloat:
        raise Exception("cos(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.cos(a0[i])
    elif a0.IsComplex:
        s, c = _dr.sincos(a0.real)
        sh, ch = _dr.sincosh(a0.imag)
        ar.real = c * ch
        ar.imag = -s * sh
    else:
        raise Exception("cos(): unsupported array type!")

    return ar


def sincos_(a0):
    if not a0.IsFloat:
        raise Exception("sincos(): requires floating point operands!")
    ar0, sr0 = _check1(a0)
    ar1 = a0.empty_(sr0 if a0.Size == Dynamic else 0)
    if not a0.IsSpecial:
        for i in range(sr0):
            result = _dr.sincos(a0[i])
            ar0[i] = result[0]
            ar1[i] = result[1]
    elif a0.IsComplex:
        s, c = _dr.sincos(a0.real)
        sh, ch = _dr.sincosh(a0.imag)
        ar0.real = s * ch
        ar0.imag = c * sh
        ar1.real = c * ch
        ar1.imag = -s * sh
    else:
        raise Exception("sincos(): unsupported array type!")
    return ar0, ar1


def csc_(a0):
    if not a0.IsFloat:
        raise Exception("csc(): requires floating point operands!")
    if not a0.IsSpecial:
        ar, sr = _check1(a0)
        for i in range(sr):
            ar[i] = _dr.csc(a0[i])
    else:
        return 1 / _dr.sin(a0)
    return ar


def sec_(a0):
    if not a0.IsFloat:
        raise Exception("sec(): requires floating point operands!")
    if not a0.IsSpecial:
        ar, sr = _check1(a0)
        for i in range(sr):
            ar[i] = _dr.sec(a0[i])
    else:
        return 1 / _dr.cos(a0)
    return ar


def tan_(a0):
    if not a0.IsFloat:
        raise Exception("tan(): requires floating point operands!")
    if not a0.IsSpecial:
        ar, sr = _check1(a0)
        for i in range(sr):
            ar[i] = _dr.tan(a0[i])
    elif a0.IsComplex:
        s, c = _dr.sincos(a0)
        return s / c
    else:
        raise Exception("tan(): unsupported array type!")
    return ar


def cot_(a0):
    if not a0.IsFloat:
        raise Exception("cot(): requires floating point operands!")
    if not a0.IsSpecial:
        ar, sr = _check1(a0)
        for i in range(sr):
            ar[i] = _dr.cot(a0[i])
    elif a0.IsComplex:
        s, c = _dr.sincos(a0)
        return c / s
    else:
        raise Exception("cot(): unsupported array type!")
    return ar


def asin_(a0):
    if not a0.IsFloat:
        raise Exception("asin(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.asin(a0[i])
    elif a0.IsSpecial:
        tmp = _dr.log(type(a0)(-a0.imag, a0.real) + _dr.sqrt(1 - _dr.sqr(a0)))
        ar.real = tmp.imag
        ar.imag = -tmp.real
    else:
        raise Exception("asin(): unsupported array type!")
    return ar


def acos_(a0):
    if not a0.IsFloat:
        raise Exception("acos(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.acos(a0[i])
    elif a0.IsSpecial:
        tmp = _dr.sqrt(1 - _dr.sqr(a0))
        tmp = _dr.log(a0 + type(a0)(-tmp.imag, tmp.real))
        ar.real = tmp.imag
        ar.imag = -tmp.real
    else:
        raise Exception("acos(): unsupported array type!")
    return ar


def atan_(a0):
    if not a0.IsFloat:
        raise Exception("atan(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.atan(a0[i])
    elif a0.IsSpecial:
        im = type(a0)(0, 1)
        tmp = _dr.log((im - a0) / (im + a0))
        return type(a0)(tmp.imag * .5, -tmp.real * 0.5)
    else:
        raise Exception("atan(): unsupported array type!")
    return ar


def atan2_(a0, a1):
    if not a0.IsFloat:
        raise Exception("atan2(): requires floating point operands!")
    ar, sr = _check2(a0, a1)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.atan2(a0[i], a1[i])
    else:
        raise Exception("atan2(): unsupported array type!")
    return ar


def exp_(a0):
    if not a0.IsFloat:
        raise Exception("exp(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.exp(a0[i])
    elif a0.IsComplex:
        s, c = _dr.sincos(a0.imag)
        exp_r = _dr.exp(a0.real)
        ar.real = exp_r * c
        ar.imag = exp_r * s
    elif a0.IsQuaternion:
        qi = a0.imag
        ri = _dr.norm(qi)
        exp_w = _dr.exp(a0.real)
        s, c = _dr.sincos(ri)
        ar.imag = qi * (s * exp_w / ri)
        ar.real = c * exp_w
    else:
        raise Exception("exp(): unsupported array type!")
    return ar


def exp2_(a0):
    if not a0.IsFloat:
        raise Exception("exp2(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.exp2(a0[i])
    elif a0.IsComplex:
        s, c = _dr.sincos(a0.imag * _dr.log_two)
        exp_r = _dr.exp2(a0.real)
        ar.real = exp_r * c
        ar.imag = exp_r * s
    else:
        raise Exception("exp2(): unsupported array type!")
    return ar


def log_(a0):
    if not a0.IsFloat:
        raise Exception("log(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.log(a0[i])
    elif a0.IsComplex:
        ar.real = .5 * _dr.log(_dr.squared_norm(a0))
        ar.imag = _dr.arg(a0)
    elif a0.IsQuaternion:
        qi_n = _dr.normalize(a0.imag)
        rq = _dr.norm(a0)
        acos_rq = _dr.acos(a0.real / rq)
        log_rq = _dr.log(rq)
        ar.imag = qi_n * acos_rq
        ar.real = log_rq
    else:
        raise Exception("log(): unsupported array type!")
    return ar


def log2_(a0):
    if not a0.IsFloat:
        raise Exception("log2(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.log2(a0[i])
    elif a0.IsComplex:
        ar.real = .5 * _dr.log2(_dr.squared_norm(a0))
        ar.imag = _dr.arg(a0) * _dr.inv_log_two
    else:
        raise Exception("log2(): unsupported array type!")
    return ar


def power_(a0, a1):
    if not a0.IsFloat:
        raise Exception("power(): requires floating point operands!")
    if not a0.IsSpecial:
        if isinstance(a1, int) or isinstance(a1, float):
            ar, sr = _check1(a0)
            for i in range(sr):
                ar[i] = _dr.power(a0[i], a1)
        else:
            ar, sr = _check2(a0, a1)
            for i in range(sr):
                ar[i] = _dr.power(a0[i], a1[i])
    else:
        return _dr.exp(_dr.log(a0) * a1)
    return ar


def sinh_(a0):
    if not a0.IsFloat:
        raise Exception("sinh(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.sinh(a0[i])
    elif a0.IsComplex:
        s, c = _dr.sincos(a0.imag)
        sh, ch = _dr.sincosh(a0.real)
        ar.real = sh * c
        ar.imag = ch * s
    else:
        raise Exception("sinh(): unsupported array type!")
    return ar


def cosh_(a0):
    if not a0.IsFloat:
        raise Exception("cosh(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.cosh(a0[i])
    elif a0.IsComplex:
        s, c = _dr.sincos(a0.imag)
        sh, ch = _dr.sincosh(a0.real)
        ar.real = ch * c
        ar.imag = sh * s
    else:
        raise Exception("cosh(): unsupported array type!")
    return ar


def sincosh_(a0):
    if not a0.IsFloat:
        raise Exception("sincosh(): requires floating point operands!")
    ar0, sr0 = _check1(a0)
    ar1 = a0.empty_(sr0 if a0.Size == Dynamic else 0)
    if not a0.IsSpecial:
        for i in range(sr0):
            result = _dr.sincosh(a0[i])
            ar0[i] = result[0]
            ar1[i] = result[1]
    elif a0.IsComplex:
        s, c = _dr.sincos(a0.imag)
        sh, ch = _dr.sincosh(a0.real)
        ar0.real = sh * c
        ar0.imag = ch * s
        ar1.real = ch * c
        ar1.imag = sh * s
    else:
        raise Exception("sincosh(): unsupported array type!")
    return ar0, ar1


def asinh_(a0):
    if not a0.IsFloat:
        raise Exception("asinh(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.asinh(a0[i])
    elif a0.IsComplex:
        return _dr.log(a0 + _dr.sqrt(_dr.sqr(a0) + 1))
    else:
        raise Exception("asinh(): unsupported array type!")
    return ar


def acosh_(a0):
    if not a0.IsFloat:
        raise Exception("acosh(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.acosh(a0[i])
    elif a0.IsComplex:
        return 2 * _dr.log(_dr.sqrt(.5 * (a0 + 1)) + _dr.sqrt(.5 * (a0 - 1)))
    else:
        raise Exception("acosh(): unsupported array type!")
    return ar


def atanh_(a0):
    if not a0.IsFloat:
        raise Exception("atanh(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.atanh(a0[i])
    elif a0.IsComplex:
        return _dr.log((1 + a0) / (1 - a0)) * .5
    else:
        raise Exception("atanh(): unsupported array type!")
    return ar


def cbrt_(a0):
    if not a0.IsFloat:
        raise Exception("cbrt(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.cbrt(a0[i])
    else:
        raise Exception("cbrt(): unsupported array type!")
    return ar


def erf_(a0):
    if not a0.IsFloat:
        raise Exception("erf(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.erf(a0[i])
    else:
        raise Exception("erf(): unsupported array type!")
    return ar


def erfinv_(a0):
    if not a0.IsFloat:
        raise Exception("erfinv(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.erfinv(a0[i])
    else:
        raise Exception("erfinv(): unsupported array type!")
    return ar


def lgamma_(a0):
    if not a0.IsFloat:
        raise Exception("lgamma(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.lgamma(a0[i])
    else:
        raise Exception("lgamma(): unsupported array type!")
    return ar


def tgamma_(a0):
    if not a0.IsFloat:
        raise Exception("tgamma(): requires floating point operands!")
    ar, sr = _check1(a0)
    if not a0.IsSpecial:
        for i in range(sr):
            ar[i] = _dr.tgamma(a0[i])
    else:
        raise Exception("tgamma(): unsupported array type!")
    return ar


# -------------------------------------------------------------------
#                       Horizontal operations
# -------------------------------------------------------------------


def all_(a0):
    if a0.IsTensor:
        return a0.array.all_()
    size = len(a0)
    if size == 0:
        return True
    value = a0[0]
    for i in range(1, size):
        value = value & a0[i]
    return value


def any_(a0):
    if a0.IsTensor:
        return a0.array.any_()
    size = len(a0)
    if size == 0:
        return False
    value = a0[0]
    for i in range(1, size):
        value = value | a0[i]
    return value


def sum_(a0):
    if a0.IsTensor:
        return a0.array.sum_()
    size = len(a0)
    if size == 0:
        return 0
    value = a0[0]
    for i in range(1, size):
        value = value + a0[i]
    return value


def prod_(a0):
    if a0.IsTensor:
        return a0.array.prod_()
    size = len(a0)
    if size == 0:
        return 1
    value = a0[0]
    for i in range(1, size):
        value = value * a0[i]
    return value


def min_(a0):
    if a0.IsTensor:
        return a0.array.min_()
    size = len(a0)
    if size == 0:
        raise Exception("min(): zero-sized array!")

    value = a0[0]
    for i in range(1, size):
        value = _dr.minimum(value, a0[i])
    return value


def max_(a0):
    if a0.IsTensor:
        return a0.array.max_()

    size = len(a0)
    if size == 0:
        raise Exception("max(): zero-sized array!")

    value = a0[0]
    for i in range(1, size):
        value = _dr.maximum(value, a0[i])
    return value


def dot_(a0, a1):
    size = len(a0)
    if size == 0:
        return 0
    if size != len(a1):
        raise Exception("dot(): incompatible array sizes!")
    if type(a0) is not type(a1):
        raise Exception("Type mismatch!")

    value = a0[0] * a1[0]
    if a0.IsFloat:
        for i in range(1, size):
            value = _dr.fma(a0[i], a1[i], value)
    else:
        for i in range(1, size):
            value += a0[i] * a1[i]
    return value


def block_sum_(a0, block_size):
    if not a0.IsArithmetic:
        raise Exception("block_sum(): requires arithmetic operands!")
    ar, sr = _check1(a0)
    for i in range(sr):
        ar[i] = _dr.block_sum(a0[i], block_size)
    return ar

# -------------------------------------------------------------------
#                     Automatic differentiation
# -------------------------------------------------------------------


def detach_(a):
    if not a.IsDiff:
        return a

    t = _dr.detached_t(type(a))

    if a.IsTensor:
        return t(a.array.detach_(), a.shape)
    else:
        result = t.empty_(len(a) if a.Size == Dynamic else 0)
        for i in range(len(a)):
            result[i] = a[i].detach_()
        return result


def grad_(a):
    if not a.IsDiff:
        return None

    t = _dr.detached_t(type(a))

    if a.IsTensor:
        return t(a.array.grad_(), a.shape)
    else:
        result = t.empty_(len(a) if a.Size == Dynamic else 0)
        for i in range(len(a)):
            result[i] = a[i].grad_()

    return result


def grad_enabled_(a):
    if not a.IsDiff:
        raise Exception("Expected a differentiable array type!")
    if a.IsTensor:
        return a.array.grad_enabled_()
    elif a.IsFloat:
        enabled = False
        for i in range(len(a)):
            # ek.Loop requires entry_ref_ here to avoid creating copies
            # of variables, whose entries are temporarily invalid
            enabled |= a.entry_ref_(i).grad_enabled_()
        return enabled
    else:
        return False


def set_grad_enabled_(a, value):
    if not a.IsDiff:
        raise Exception("Expected a differentiable array type!")
    if a.IsTensor:
        a.array.set_grad_enabled_(value)
    else:
        for i in range(len(a)):
            a.entry_ref_(i).set_grad_enabled_(value)


def set_grad_(a, grad):
    if not a.IsDiff:
        raise Exception("Expected a differentiable array type!")

    if a.IsTensor:
        if _dr.is_tensor_v(grad):
            a.array.set_grad_(grad.array)
        else:
            a.array.set_grad_(grad)
    else:
        s = len(a)
        for i in range(s):
            a[i].set_grad_(grad[i])


def accum_grad_(a, grad):
    if not a.IsDiff:
        raise Exception("Expected a differentiable array type!")

    if a.IsTensor:
        if _dr.is_tensor_v(grad):
            a.array.accum_grad_(grad.array)
        else:
            a.array.accum_grad_(grad)
    else:
        s = len(a)
        for i in range(s):
            a[i].accum_grad_(grad[i])


def enqueue_(a, mode):
    if not a.IsDiff:
        raise Exception("Expected a differentiable array type!")
    if a.IsTensor:
        a.array.enqueue_(mode)
    else:
        for i in range(len(a)):
            a[i].enqueue_(mode)


def migrate_(a, target):
    if not a.IsJIT:
        raise Exception("Expected a JIT array type!")
    t = type(a)
    if a.IsTensor:
        return t(a.array.migrate_(target), a.shape)
    result = t.empty_(len(a) if a.Size == Dynamic else 0)
    for i in range(len(a)):
        result[i] = a[i].migrate_(target)
    return result


def index_(a):
    if not a.IsJIT:
        raise Exception("Expected a JIT array type!")
    return tuple(v.index for v in a)


def index_ad_(a):
    if not a.IsDiff:
        raise Exception("Expected a differentiable array type!")
    return tuple(v.index_ad for v in a)


# -------------------------------------------------------------------
#                      Initialization operations
# -------------------------------------------------------------------

def assign(self, other):
    if self is other:
        return
    elif len(self) != len(other):
        raise Exception("assign(): size mismatch!")
    else:
        for i in range(len(self)):
            self[i] = other[i]


def broadcast_(self, value):
    if not self.IsSpecial:
        for i in range(len(self)):
            self.set_entry_(i, value)
    elif self.IsComplex:
        self.set_entry_(0, value)
        self.set_entry_(1, 0)
    elif self.IsQuaternion:
        for i in range(3):
            self.set_entry_(i, 0)
        self.set_entry_(3, value)
    elif self.IsMatrix:
        t = self.Value
        for i in range(len(self)):
            c = _dr.zeros(t)
            c.set_entry_(i, t.Value(value))
            self.set_entry_(i, c)
    else:
        raise Exception("broadcast_(): don't know how to handle this type!")


@classmethod
def empty_(cls, shape):
    if cls.IsTensor:
        shape = [shape] if type(shape) is int else shape
        return cls(_dr.empty(cls.Array, _dr.prod(shape)), shape)

    result = cls()
    if cls.Size == Dynamic:
        result.init_(shape)
    elif cls.IsDynamic:
        for i in range(len(result)):
            result.set_entry_(i, _dr.empty(cls.Value, shape))
    return result


@classmethod
def zero_(cls, shape=1):
    if cls.IsTensor:
        shape = [shape] if type(shape) is int else shape
        return cls(_dr.zeros(cls.Array, _dr.prod(shape)), shape)

    result = cls()
    if cls.Size == Dynamic:
        result.init_(shape)
        for i in range(shape):
            result.set_entry_(i, 0)
    else:
        for i in range(cls.Size):
            result.set_entry_(i, _dr.zeros(cls.Value, shape))
    return result


@classmethod
def full_(cls, value, shape):
    if cls.IsTensor:
        shape = [shape] if type(shape) is int else shape
        return cls(_dr.full(cls.Array, value, _dr.prod(shape)), shape)

    result = cls()
    if cls.Size == Dynamic:
        result.init_(shape)
        for i in range(shape):
            result.set_entry_(i, value)
    else:
        if _dr.depth_v(value) != cls.Depth - 1:
            value = _dr.full(cls.Value, value, shape)

        for i in range(cls.Size):
            result.set_entry_(i, value)
    return result


@classmethod
def linspace_(cls, min, max, size=1, endpoint=True):
    if cls.IsTensor:
        raise Exception("linspace_(): Tensor type not supported!")
    result = cls.empty_(size)
    step = (max - min) / (len(result) - (1 if endpoint and (size > 1) else 0))
    if cls.IsFloat:
        for i in range(len(result)):
            result[i] = min + step * i
    else:
        for i in range(len(result)):
            result[i] = _dr.fma(step, i, min)
    return result


@classmethod
def arange_(cls, start, end, step):
    if cls.IsTensor:
        raise Exception("arange_(): Tensor type not supported!")
    size = (end - start + step - (1 if step > 0 else -1)) / step
    result = cls.empty_(size)
    for i in range(len(result)):
        result[i] = start + step*i
    return result


@classmethod
def gather_(cls, source, index, mask):
    assert source.Depth == 1
    sr = max(len(index), len(mask))
    result = cls.empty_(sr if cls.Size == Dynamic else 0)
    for i in range(sr):
        result[i] = _dr.gather(cls.Value, source, index[i], mask[i])
    return result


def scatter_(self, target, index, mask):
    assert target.Depth == 1
    sr = max(len(self), len(index), len(mask))
    for i in range(sr):
        _dr.scatter(target, self[i], index[i], mask[i])


def scatter_reduce_(self, op, target, index, mask):
    assert target.Depth == 1
    sr = max(len(self), len(index), len(mask))
    for i in range(sr):
        _dr.scatter_reduce(op, target, self[i], index[i], mask[i])


# -------------------------------------------------------------------
#    Interoperability with other frameworks (NumPy, PyTorch, Jax)
# -------------------------------------------------------------------


def export_(a, migrate_to_host, version, owner_supported=True):
    shape = _dr.shape(a)
    ndim = len(shape)
    if not a.IsTensor:
        shape = tuple(reversed(shape))

    if not a.IsJIT:
        # F-style strides
        temp, strides = a.Type.Size, [0] * ndim

        if a.IsTensor:
            # First dimension is the dynamic one, the rest should be in reversed order
            for i in reversed(range(1, ndim)):
                strides[i] = temp
                temp *= shape[i]
            strides[0] = temp
        else:
            # Dr.Jit represents 3D arrays as 4D to leverage SIMD instructions
            padding = 1 if a.IsScalar and a.IsMatrix and shape[0] == 3 else 0
            for i in range(ndim):
                strides[i] = temp
                temp *= shape[i] + padding

        # Array is already contiguous in memory -- document its structure
        return {
            "shape": shape,
            "strides": tuple(strides),
            "typestr": "<" + a.Type.NumPy,
            "data": (a.data_(), False),
            "version": version,
            "device": -1,
            "owner": a
        }
    else:
        # C-style strides
        temp, strides = a.Type.Size, [0] * ndim

        # First dimension is the dynamic one, the rest should be in reversed order
        for i in reversed(range(1, ndim)):
            strides[i if a.IsTensor else (ndim - i)] = temp
            temp *= shape[i]
        strides[0] = temp

        # JIT array -- requires extra transformations
        b = _dr.ravel(_dr.detach(a) if a.IsDiff else a)
        _dr.eval(b)

        if b.IsCUDA and migrate_to_host:
            if b is a:
                b = type(a)(b)
            b = b.migrate_(_dr.AllocType.Host)
            _dr.sync_thread()
        elif b.IsLLVM:
            _dr.sync_thread()

        if not owner_supported and a is not b:
            # If the caller cannot deal with the 'owner' field, use
            # a weak reference to keep 'b' alive while 'a' exists
            _wr.finalize(a, lambda arg: None, b)

        data_ptr = b.data_()
        if data_ptr == 0:
            # NumPy does not accept null pointers (even when the array is empty)
            data_ptr = 1

        record = {
            "shape": shape,
            "strides": tuple(strides),
            "typestr": "<" + a.Type.NumPy,
            "data": (data_ptr, False),
            "version": version,
            "device": _dr.device(b),
            "owner": b
        }

        return record


@property
def op_array_interface(a):
    try:
        return a.export_(migrate_to_host=True, version=3,
                         owner_supported=False)
    except BaseException as e:
        print(e)


@property
def op_cuda_array_interface(a):
    if not a.IsCUDA:
        raise Exception("__cuda_array_interface__: only CUDA "
                        "arrays are supported!")
    return a.export_(migrate_to_host=False, version=2)


def numpy(a):
    import numpy
    arr = numpy.array(a, copy=False)
    if a.IsComplex:
        arr = arr.ravel()
        if arr.dtype == numpy.float32:
            return arr.view(numpy.complex64)[...]
        elif arr.dtype == numpy.float64:
            return arr.view(numpy.complex128)[...]
        else:
            raise Exception("Unsupported dtype for complex conversion!")
    return arr


def op_dlpack(a):
    struct = a.export_(migrate_to_host=False, version=2)
    isize = a.Type.Size
    strides = tuple(k // isize for k in struct["strides"])
    return _dr.detail.to_dlpack(
        owner=struct["owner"],
        data=struct["data"][0],
        type=a.Type,
        device=struct["device"],
        shape=struct["shape"],
        strides=strides
    )


def torch(a):
    from torch.utils.dlpack import from_dlpack
    return from_dlpack(a.__dlpack__())


def jax(a):
    from jax.dlpack import from_dlpack
    from jax import devices
    if a.IsLLVM:
        try:
            # Not all Jax versions accept the 'backend' parameter
            return from_dlpack(a.__dlpack__(), backend=devices(backend="cpu")[0])
        except:
            pass
    return from_dlpack(a.__dlpack__())


def tf(a):
    from tensorflow.experimental.dlpack import from_dlpack
    from tensorflow import constant
    constant(0) # Dummy op to ensure that the Tensorflow context is initialized
    return from_dlpack(a.__dlpack__())


def op_iter(a):
    size = len(a)
    class array_iterator:
        def __init__(self):
            self.i = 0

        def __iter__(self):
            return self

        def __next__(self):
            if self.i >= size:
                raise StopIteration()
            self.i += 1
            return a[self.i-1]
    return array_iterator()


import drjit as dr
import pytest
import importlib

from drjit.scalar import Matrix2f as M2
from drjit.scalar import Matrix3f as M3
from drjit.scalar import Matrix4f as M4
from drjit.scalar import Float
M = M4

def prepare(pkg):
    if 'cuda' in pkg:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    elif 'llvm' in pkg:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')
    return importlib.import_module(pkg)


def test01_init_broadcast_mul():
    m = M(*range(1, 17))
    m2 = m @ m + 1
    m2[1, 2] = 3
    assert m2[1, 2] == 3

    assert m2 == M(
        91, 100, 110, 120,
        202, 229, 3, 280,
        314, 356, 399, 440,
        426, 484, 542, 601
    )

    assert m2[0] == m2.Value(91, 202, 314, 426)


def test02_transpose_diag():
    m = dr.transpose(M(*range(1, 17)))
    assert m == M(
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [9, 10, 11, 12],
        [13, 14, 15, 16]
    )

    assert dr.diag(m) == [1, 6, 11, 16]
    assert dr.diag(dr.diag(m)) == M(
        [1, 0, 0, 0],
        [0, 6, 0, 0],
        [0, 0, 11, 0],
        [0, 0, 0, 16]
    )


def test03_roundtrip():
    pytest.importorskip("numpy")
    m = M(*range(1, 17)) + dr.full(M, dr.arange(dr.scalar.Float))
    m2 = M(m.numpy())
    assert m == m2


def test04_trace_frob():
    m = M(*range(1, 17))
    assert dr.trace(m) == 34
    assert dr.frob(m) == 1496


def test05_allclose():
    m = dr.full(M, 1)
    assert dr.allclose(m, dr.full(M, 1))
    m[1, 0] = 0
    assert not dr.allclose(m, dr.full(M, 1))
    m = dr.identity(M)
    assert dr.allclose(m, 1)
    m[1, 0] = 1
    assert not dr.allclose(m, 1)


@pytest.mark.parametrize('M', [M2, M3, M4])
def test06_det(M):
    np = pytest.importorskip("numpy")

    np.random.seed(1)
    for i in range(100):
        m1 = np.float32(np.random.normal(size=list(reversed(M.Shape))))
        m2 = M(m1)
        det1 = Float(np.linalg.det(m1))
        det2 = dr.det(m2)
        assert dr.allclose(det1, det2, atol=1e-6)


@pytest.mark.parametrize('M', [M2, M3, M4])
def test07_inverse(M):
    np = pytest.importorskip("numpy")

    np.random.seed(1)
    for i in range(100):
        m1 = np.float32(np.random.normal(size=list(reversed(M.Shape))))
        m2 = M(m1)
        inv1 = M(np.linalg.inv(m1))@m2 - dr.identity(M)
        inv2 = dr.inverse(m2)@m2 - dr.identity(M)
        assert dr.allclose(inv1, 0, atol=1e-3)
        assert dr.allclose(inv2, 0, atol=1e-3)


def test08_polar():
    m = M(*range(1, 17)) + dr.identity(M)
    q, r = dr.polar_decomp(m)
    assert dr.allclose(q@r, m)
    assert dr.allclose(q@dr.transpose(q), dr.identity(M), atol=1e-6)


def test09_transform_decompose():
    m = dr.scalar.Matrix4f([[1, 0, 0, 8], [0, 2, 0, 7], [0, 0, 9, 6], [0, 0, 0, 1]])
    s, q, t = dr.transform_decompose(m)

    assert dr.allclose(s, dr.scalar.Matrix3f(m))
    assert dr.allclose(q, dr.scalar.Quaternion4f(1))
    assert dr.allclose(t, [8, 7, 6])
    assert dr.allclose(m, dr.transform_compose(s, q, t))

    q2 = dr.rotate(dr.scalar.Quaternion4f, dr.scalar.Array3f(0, 0, 1), 15.0)
    m @= dr.quat_to_matrix(q2)
    s, q, t = dr.transform_decompose(m)

    assert dr.allclose(q, q2)


def test10_matrix_to_quat():
    q = dr.rotate(dr.scalar.Quaternion4f, dr.scalar.Array3f(0, 0, 1), 15.0)
    m = dr.quat_to_matrix(q)
    q2 = dr.matrix_to_quat(m)
    assert dr.allclose(q, q2)


@pytest.mark.parametrize("package", ["drjit.scalar", "drjit.cuda", "drjit.llvm"])
def test11_constructor(package):
    """
    Check Matrix construction from Python array and Numpy array
    """
    np = pytest.importorskip("numpy")

    package = prepare(package)
    Float, Matrix3f = package.Float, package.Matrix3f

    m1 = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]]
    m2 = np.array(m1, dtype=np.float32)
    m3 = Matrix3f(m1)
    m4 = Matrix3f(m2)

    assert dr.allclose(m3, m1)
    assert dr.allclose(m3, m2)
    assert dr.allclose(m3, m4)
    assert dr.allclose(m4, m2)

    if dr.is_jit_v(Float):
        np.random.seed(1)
        for i in range(1, 4):
            values = np.random.random((i, 3, 3)).astype('float32')
            m5 = Matrix3f(values)
            assert dr.allclose(m5, values)


@pytest.mark.parametrize("package", ["drjit.scalar", "drjit.cuda", "drjit.llvm"])
def test12_matrix_scale(package):
    np = pytest.importorskip("numpy")

    package = prepare(package)
    Float, Matrix3f = package.Float, package.Matrix3f

    m = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]], dtype=np.float32)
    m2 = np.float32(2*m)

    assert package.Matrix3f(m) * 2 == package.Matrix3f(m2)
    assert package.Matrix3f(m) @ 2 == package.Matrix3f(m2)
    assert package.Matrix3f(m) * package.Float(2) == package.Matrix3f(m2)
    assert package.Matrix3f(m) @ package.Float(2) == package.Matrix3f(m2)
    assert 2 * package.Matrix3f(m) == package.Matrix3f(m2)
    assert package.Float(2) * package.Matrix3f(m) == package.Matrix3f(m2)


@pytest.mark.parametrize("package", ["drjit.scalar", "drjit.cuda", "drjit.llvm"])
def test12_matrix_vector(package):
    np = pytest.importorskip("numpy")

    m_ = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]], dtype=np.float32)
    package = prepare(package)
    Float, Matrix3f, Array3f = package.Float, package.Matrix3f, package.Array3f
    m = Matrix3f(m_)
    v1 = m @ Array3f(1, 0, 0)
    v2 = m @ Array3f(1, 1, 0)
    assert dr.allclose(v1, [0.1, 0.4, 0.7])
    assert dr.allclose(v2, [0.3, 0.9, 1.5])
    v1 = m @ dr.scalar.Array3f(1, 0, 0)
    v2 = m @ dr.scalar.Array3f(1, 1, 0)
    assert dr.allclose(v1, [0.1, 0.4, 0.7])
    assert dr.allclose(v2, [0.3, 0.9, 1.5])

    with pytest.raises(dr.Exception):
        m * Array3f(1, 0, 0)

    with pytest.raises(dr.Exception):
        m * dr.scalar.Array3f(1, 0, 0)

    with pytest.raises(dr.Exception):
        m * m

    with pytest.raises(dr.Exception):
        m * dr.scalar.Matrix3f(m_)


@pytest.mark.parametrize("package", ["drjit.scalar", "drjit.cuda", "drjit.llvm"])
def test13_matmul_other(package):
    package = prepare(package)
    Float, Matrix3f, Array3f = package.Float, package.Matrix3f, package.Array3f

    m = Matrix3f(1,2,3,4,5,6,7,8,9)
    m2 = Matrix3f(30, 36, 42, 66, 81, 96, 102, 126, 150)
    v = Array3f(5,2,1)
    assert dr.allclose(m @ m, m2)
    assert dr.allclose(m @ v, Array3f(12, 36, 60))
    assert dr.allclose(v @ m, Array3f(20, 28, 36))
    assert dr.allclose(v @ v, Float(30))


@pytest.mark.parametrize("package", ["drjit.cuda", "drjit.llvm"])
@pytest.mark.parametrize("dest", ['numpy', 'torch', 'jax'])
def test14_roundtrip(package, dest):
    pytest.importorskip(dest)

    if dest == 'jax' and 'cuda' in package:
        try:
            from jax.lib import xla_bridge
            xla_bridge.get_backend('gpu')
        except:
            pytest.skip('Backend gpu failed to initialize on JAX')

    package = prepare(package)
    Float, Array3f, Array4f = package.Float, package.Array3f, package.Array4f
    Matrix3f, Matrix4f, Matrix44f = package.Matrix3f, package.Matrix4f, package.Matrix44f

    def to_dest(a):
        if dest == 'numpy':
            return a.numpy()
        if dest == 'torch':
            return a.torch()
        if dest == 'jax':
            return a.jax()

    v = Array3f(
        (1.0 + dr.arange(Float, 5)),
        (1.0 + dr.arange(Float, 5)) * 2,
        (1.0 + dr.arange(Float, 5)) * 3,
    )
    assert(v == Array3f(to_dest(v)))

    m = Matrix3f([
        Array3f([[1.0, 1.5], [2.0, 2.5], [3.0, 3.5]]),
        Array3f([[1.0, 1.5], [2.0, 2.5], [3.0, 3.5]]) * 10,
        Array3f([[1.0, 1.5], [2.0, 2.5], [3.0, 3.5]]) * 100
    ])
    assert(m == Matrix3f(to_dest(m)))

    m = Matrix4f([
        Array4f([[1.0, 1.5], [2.0, 2.5], [3.0, 3.5], [4.0, 4.5]]),
        Array4f([[1.0, 1.5], [2.0, 2.5], [3.0, 3.5], [4.0, 4.5]]) * 10,
        Array4f([[1.0, 1.5], [2.0, 2.5], [3.0, 3.5], [4.0, 4.5]]) * 100,
        Array4f([[1.0, 1.5], [2.0, 2.5], [3.0, 3.5], [4.0, 4.5]]) * 1000
    ])
    assert(m == Matrix4f(to_dest(m)))

    m = Matrix44f([
        [1.0, 2.0, 3.0, 4.0],
        [1.0, 2.0, 3.0, 4.0],
        [1.0, 2.0, 3.0, 4.0],
        [1.0, 2.0, 3.0, 4.0]
    ])
    assert(m == Matrix44f(to_dest(m)))


@pytest.mark.parametrize("package", ["drjit.scalar"])
def test15_quat_to_euler(package):
    np = pytest.importorskip("numpy")

    package = prepare(package)
    Quaternion4f, Array3f, Float = package.Quaternion4f, package.Array3f, package.Float

    # Gimbal lock at +pi/2
    q = Quaternion4f(0, 1.0 / np.sqrt(2), 0, 1.0 / np.sqrt(2))
    assert(dr.allclose(dr.quat_to_euler(q), Array3f(0, np.pi / 2, 0)))
    # Gimbal lock at -pi/2
    q = Quaternion4f(0, -1.0 / np.sqrt(2), 0, 1.0 / np.sqrt(2))
    assert(dr.allclose(dr.quat_to_euler(q), Array3f(0, -np.pi / 2, 0)))
    # Gimbal lock at +pi/2, such that computed sinp > 1
    q = Quaternion4f(0, 1.0 / np.sqrt(2) + 1e-6, 0, 1.0 / np.sqrt(2))
    assert(dr.allclose(dr.quat_to_euler(q), Array3f(0, np.pi / 2, 0)))
    # Gimbal lock at -pi/2, such that computed sinp < -1
    q = Quaternion4f(0, -1.0 / np.sqrt(2) - 1e-6, 0, 1.0 / np.sqrt(2))
    assert(dr.allclose(dr.quat_to_euler(q), Array3f(0, -np.pi / 2, 0)))
    # Quaternion without gimbal lock
    q = Quaternion4f(0.15849363803863525, 0.5915063619613647, 0.15849363803863525, 0.7745190262794495)
    e = Array3f(np.pi / 3, np.pi / 3, np.pi / 3)
    assert(dr.allclose(dr.quat_to_euler(q), e))
    # Round trip
    assert(dr.allclose(e, dr.quat_to_euler(dr.euler_to_quat(e))))
    # Euler -> Quat
    assert(dr.allclose(q, dr.euler_to_quat(e)))



import drjit as dr
import pytest
import gc

def get_class(name):
    """Resolve a package+class name into the corresponding type"""
    if 'cuda' in name:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    elif 'llvm' in name:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')

    name = name.split('.')
    value = __import__(".".join(name[:-1]))
    for item in name[1:]:
        value = getattr(value, item)
    dr.set_flag(dr.JitFlag.LoopRecord, True)

    return value

def setup_function(function):
    dr.set_flag(dr.JitFlag.LoopRecord, True)

def teardown_function(function):
    dr.set_flag(dr.JitFlag.LoopRecord, False)

pkgs = ["drjit.cuda", "drjit.cuda.ad",
        "drjit.llvm", "drjit.llvm.ad"]

pkgs_ad = ["drjit.cuda.ad",
           "drjit.llvm.ad"]

@pytest.mark.parametrize("pkg", pkgs)
def test01_ctr(pkg):
    p = get_class(pkg)

    i = dr.arange(p.Int, 0, 10)

    loop = p.Loop("MyLoop", lambda: i)
    while loop(i < 5):
        i = i + 1

    assert i == p.Int(5, 5, 5, 5, 5, 5, 6, 7, 8, 9)


@pytest.mark.parametrize("pkg", pkgs)
def test02_record_loop(pkg):
    p = get_class(pkg)

    for i in range(3):
        dr.set_flag(dr.JitFlag.LoopRecord, not i == 0)
        dr.set_flag(dr.JitFlag.LoopOptimize, i == 2)

        for j in range(2):
            x = dr.arange(p.Int, 0, 10)
            y = dr.zeros(p.Float, 1)
            z = p.Float(1)

            loop = p.Loop("MyLoop", lambda: (x, y, z))
            while loop(x < 5):
                y += p.Float(x)
                x += 1
                z = z + 1

            if j == 0:
                dr.schedule(x, y, z)

            assert z == p.Int(6, 5, 4, 3, 2, 1, 1, 1, 1, 1)
            assert y == p.Int(10, 10, 9, 7, 4, 0, 0, 0, 0, 0)
            assert x == p.Int(5, 5, 5, 5, 5, 5, 6, 7, 8, 9)


@pytest.mark.parametrize("variant", [0, 1])
@pytest.mark.parametrize("pkg", pkgs)
def test03_multiple_values(pkg, variant):
    p = get_class(pkg)

    i = dr.arange(p.Int, 0, 10)
    v = dr.zeros(p.Array3f, 10)

    if variant == 1:
        v.y = p.Float(0)

    loop = p.Loop("MyLoop", lambda: (i, v))
    while loop(i < 5):
        i.assign(i + 1)
        f = p.Float(i)
        v.x += f
        v.y += 2*f
        v.z += 4*f

    if variant == 0:
        dr.eval(i, v)
    else:
        dr.eval(i)
        dr.eval(v.x)
        dr.eval(v.y)
        dr.eval(v.z)

    assert i == p.Int(5, 5, 5, 5, 5, 5, 6, 7, 8, 9)
    assert v.y == p.Int(30, 28, 24, 18, 10, 0, 0, 0, 0, 0)


@pytest.mark.parametrize("pkg", pkgs)
def test04_side_effect(pkg):
    p = get_class(pkg)

    i = dr.zeros(p.Int, 10)
    j = dr.zeros(p.Int, 10)
    buf = dr.zeros(p.Float, 10)

    loop = p.Loop("MyLoop", lambda: (i, j))
    while loop(i < 10):
        j += i
        i += 1
        dr.scatter_reduce(op=dr.ReduceOp.Add, target=buf, value=p.Float(i), index=0)

    dr.eval(i, j)
    assert i == p.Int([10]*10)
    assert buf == p.Float(550, *([0]*9))
    assert j == p.Int([45]*10)


@pytest.mark.parametrize("pkg", pkgs)
def test05_side_effect_noloop(pkg):
    p = get_class(pkg)

    i = dr.zeros(p.Int, 10)
    j = dr.zeros(p.Int, 10)
    buf = dr.zeros(p.Float, 10)
    dr.set_flag(dr.JitFlag.LoopRecord, False)

    loop = p.Loop("MyLoop", lambda: (i, j))
    while loop(i < 10):
        j += i
        i += 1
        dr.scatter_reduce(op=dr.ReduceOp.Add, target=buf, value=p.Float(i), index=0)

    assert i == p.Int([10]*10)
    assert buf == p.Float(550, *([0]*9))
    assert j == p.Int([45]*10)


@pytest.mark.parametrize("variant", [0, 1, 2])
@pytest.mark.parametrize("pkg", pkgs)
def test06_test_collatz(pkg, variant):
    p = get_class(pkg)

    def collatz(value: p.Int):
        counter = p.Int(0)
        loop = p.Loop("collatz", lambda: (value, counter))
        while (loop(dr.neq(value, 1))):
            is_even = dr.eq(value & 1, 0)
            value.assign(dr.select(is_even, value // 2, 3*value + 1))
            counter += 1
        return value, counter

    value, ctr = collatz(dr.arange(p.Int, 1, 11))
    if variant == 0:
        dr.eval(value, ctr)
    elif variant == 1:
        dr.eval(value)
        dr.eval(ctr)
    elif variant == 2:
        dr.eval(ctr)
        dr.eval(value)

    assert value == p.Int([1]*10)
    assert ctr == p.Int([0,1,7,2,5,8,16,3,19,6])

@pytest.mark.parametrize("variant", [0, 1])
@pytest.mark.parametrize("pkg", ["drjit.cuda",
                                 "drjit.cuda.ad",
                                 "drjit.llvm",
                                 "drjit.llvm.ad"])
def test07_loop_nest(pkg, variant):
    p = get_class(pkg)

    def collatz(value: p.Int):
        counter = p.Int(0)
        loop = p.Loop("Nested", lambda: (value, counter))
        while (loop(dr.neq(value, 1))):
            is_even = dr.eq(value & 1, 0)
            value.assign(dr.select(is_even, value // 2, 3*value + 1))
            counter += 1
        return counter

    i = p.Int(1)
    buf = dr.full(p.Int, 1000, 16)
    dr.eval(buf)

    if variant == 0:
        loop_1 = p.Loop("MyLoop", lambda: i)
        while loop_1(i <= 10):
            dr.scatter(buf, collatz(p.Int(i)), i - 1)
            i += 1
    else:
        for i in range(1, 11):
            dr.scatter(buf, collatz(p.Int(i)), i - 1)
            i += 1

    assert buf == p.Int(0, 1, 7, 2, 5, 8, 16, 3, 19, 6, 1000, 1000, 1000, 1000, 1000, 1000)



@pytest.mark.parametrize("pkg", pkgs)
def test08_loop_simplification(pkg):
    # Test a regression where most loop variables are optimized away
    p = get_class(pkg)
    res = dr.zeros(p.Float, 10)
    active = p.Bool(True)
    depth = p.UInt32(1000)

    loop = p.Loop("TestLoop")
    loop.put(lambda: (active, depth, res))
    loop.init()
    while loop(active):
        res += dr.arange(p.Float, 10)
        depth -= 1
        active &= (depth > 0)
    del loop
    del active
    del depth
    assert res == dr.arange(p.Float, 10) * 1000


@pytest.mark.parametrize("pkg", pkgs)
def test09_failure_invalid_loop_arg(pkg):
    p = get_class(pkg)
    i = p.Int(1)
    with pytest.raises(RuntimeError) as e:
        p.Loop("MyLoop", i)
    assert 'expected a lambda function' in str(e.value)

    i = p.Int(1)
    with pytest.raises(TypeError) as e:
        l = p.Loop("MyLoop")
        l.put(i)
    assert 'incompatible function arguments' in str(e.value)
    del l


@pytest.mark.parametrize("pkg", pkgs_ad)
def test10_failure_invalid_type(pkg):
    p = get_class(pkg)
    i = 1
    with pytest.raises(dr.Exception) as e:
        p.Loop("MyLoop", lambda: i)
    assert 'you must use Dr.Jit arrays/structs' in str(e.value)


@pytest.mark.parametrize("variant", [0, 1])
@pytest.mark.parametrize("pkg", pkgs)
def test11_failure_unitialized(pkg, variant):
    p = get_class(pkg)
    i = p.Int(0)
    j = p.Int() if variant == 0 else p.Int(1)
    l = None
    with pytest.raises(dr.Exception) as e:
        l = p.Loop("MyLoop", lambda: (i, j))
        while l(i < 10):
            j = p.Int() if variant == 1 else i + 1
            i += 1

    assert 'is uninitialized!' in str(e.value)
    del l


@pytest.mark.parametrize("pkg", pkgs)
def test12_failure_changing_type(pkg):
    p = get_class(pkg)
    i = p.Int(0)
    with pytest.raises(dr.Exception) as e:
        l = p.Loop("MyLoop", lambda: i)
        while l(i < 10):
            i = p.Float(0)
    assert 'the type of loop state variables must remain the same throughout the loop' in str(e.value)
    del l


@pytest.mark.parametrize("variant", [0, 1])
@pytest.mark.parametrize("pkg", pkgs_ad)
def test13_failures_ad(pkg, variant):
    p = get_class(pkg)
    i = p.Int(0)
    v = p.Array3f(1,2,2)
    l = None
    with pytest.raises(dr.Exception) as e:
        if variant == 0:
            dr.enable_grad(v)
        l = p.Loop("MyLoop", lambda: (i, v))
        while l(i < 10):
            i = i + 1
            if variant == 1:
                dr.enable_grad(v)
    print(str(e.value))
    assert 'one of the supplied loop state variables of type Float is attached to the AD graph' in str(e.value)
    del l


@pytest.mark.parametrize("pkg", pkgs)
def test14_failure_state_leak(pkg):
    p = get_class(pkg)
    i, j = p.Int(0), p.Int(0)

    l = p.Loop("MyLoop", lambda: i)
    while l(i < 10):
        i = i + 1
        j += i

    with pytest.raises(RuntimeError) as e:
        j[0]
    assert 'placeholder variables are used to record computation symbolically' in str(e.value)
    del l


@pytest.mark.parametrize("pkg", pkgs)
def test15_scalar_side_effect(pkg):
    # Ensure that a scalar side effect takes place multiple times if the loop processes larger arrays
    p = get_class(pkg)

    for i in range(1):
        dr.set_flag(dr.JitFlag.LoopRecord, i == 1)

        active = p.Bool(True)
        unrelated = dr.arange(p.UInt32, 123)

        loop = p.Loop("Test", lambda: (active, unrelated))
        target = dr.zeros(p.UInt32)

        while loop(active):
            dr.scatter_reduce(dr.ReduceOp.Add, target, 1, 0)
            active &= False

        assert target[0] == 123
